"""
Combinatorial Hopf algebras

- :ref:`sage.combinat.sf.all`
- :ref:`sage.combinat.ncsf_qsym.all`
- :ref:`sage.combinat.ncsym.all`
- :ref:`sage.combinat.schubert_polynomial`
- :ref:`sage.combinat.chas.fsym`
- :ref:`sage.combinat.fqsym`
- :ref:`sage.combinat.grossman_larson_algebras`
- :ref:`sage.combinat.chas.wqsym`
"""
# install the docstring of this module to the containing package
from sage.misc.namespace_package import install_doc
install_doc(__package__, __doc__)

from sage.misc.lazy_import import lazy_import

lazy_import('sage.combinat.chas.fsym', ['FreeSymmetricFunctions'])
lazy_import('sage.combinat.chas.wqsym', ['WordQuasiSymmetricFunctions'])
del lazy_import
del install_doc



r"""
Groups of isometries

Let `M = \ZZ^n` or `\QQ^n`, `b: M \times M \rightarrow \QQ` a bilinear form and
`f: M \rightarrow M` a linear map. We say that `f` is an isometry if for all
elements `x,y` of `M` we have that `b(x,y)=b(f(x),f(y))`.
A group of isometries is a subgroup of `GL(M)` consisting of isometries.

EXAMPLES::

    sage: L = IntegralLattice("D4")                                                     # needs sage.graphs
    sage: O = L.orthogonal_group(); O                                                   # needs sage.graphs
    Group of isometries with 3 generators (
    [0 0 0 1]  [ 1  1  0  0]  [ 1  0  0  0]
    [0 1 0 0]  [ 0  0  1  0]  [-1 -1 -1 -1]
    [0 0 1 0]  [ 0  1  0  1]  [ 0  0  1  0]
    [1 0 0 0], [ 0 -1 -1  0], [ 0  0  0  1]
    )

Basic functionality is provided by GAP::

    sage: O.cardinality()                                                               # needs sage.graphs
    1152
    sage: len(O.conjugacy_classes_representatives())                                    # needs sage.graphs
    25

AUTHORS:

- Simon Brandhorst (2018-02): First created
"""

# ****************************************************************************
#       Copyright (C) 2018 Simon Brandhorst <sbrandhorst@web.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.groups.matrix_gps.finitely_generated_gap import FinitelyGeneratedMatrixGroup_gap
from sage.categories.action import Action


class GroupOfIsometries(FinitelyGeneratedMatrixGroup_gap):
    r"""
    A base class for Orthogonal matrix groups with a gap backend.

    Main difference to :class:`~sage.groups.matrix_gps.orthogonal.OrthogonalMatrixGroup_gap`
    is that we can specify generators and a bilinear form. Following GAP, the group action is
    from the right.

    INPUT:

    - ``degree`` -- integer; the degree (matrix size) of the matrix
    - ``base_ring`` -- ring, the base ring of the matrices
    - ``gens`` -- list of matrices over the base ring
    - ``invariant_bilinear_form`` -- a symmetric matrix
    - ``category`` -- (default: ``None``) a category of groups
    - ``check`` -- boolean (default: ``True``); check if the generators
      preserve the bilinear form
    - ``invariant_submodule`` -- a submodule preserved by the group action
      (default: ``None``); registers an action on this submodule
    - ``invariant_quotient_module`` -- a quotient module preserved by
      the group action (default: ``None``)
      registers an action on this quotient module

    EXAMPLES::

        sage: from sage.groups.matrix_gps.isometries import GroupOfIsometries
        sage: bil = Matrix(ZZ, 2, [3,2,2,3])
        sage: gens = [-Matrix(ZZ, 2, [0,1,1,0])]
        sage: O = GroupOfIsometries(2, ZZ, gens, bil)
        sage: O
        Group of isometries with 1 generator (
        [ 0 -1]
        [-1  0]
        )
        sage: O.order()
        2

    Infinite groups are O.K. too::

        sage: bil = Matrix(ZZ,4,[0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0])
        sage: f = Matrix(ZZ,4,[0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, -1, 1, 1, 1])
        sage: O = GroupOfIsometries(2, ZZ, [f], bil)
        sage: O.cardinality()
        +Infinity
    """

    def __init__(self, degree, base_ring,
                 gens, invariant_bilinear_form,
                 category=None, check=True,
                 invariant_submodule=None,
                 invariant_quotient_module=None):
        r"""
        Create this orthogonal group from the input.

        TESTS::

            sage: from sage.groups.matrix_gps.isometries import GroupOfIsometries
            sage: bil = Matrix(ZZ,2,[3,2,2,3])
            sage: gens = [-Matrix(ZZ,2,[0,1,1,0])]
            sage: cat = Groups().Finite()
            sage: O = GroupOfIsometries(2, ZZ, gens, bil, category=cat)
            sage: TestSuite(O).run()
        """
        from copy import copy
        G = copy(invariant_bilinear_form)
        G.set_immutable()
        self._invariant_bilinear_form = G
        self._invariant_submodule = invariant_submodule
        self._invariant_quotient_module = invariant_quotient_module
        if check:
            I = invariant_submodule
            Q = invariant_quotient_module
            for f in gens:
                self._check_matrix(f)
                if (I is not None) and I * f != I:
                    raise ValueError("the submodule is not preserved")
                if Q is not None and (Q.W() != Q.W()*f or Q.V()*f != Q.V()):
                    raise ValueError("the quotient module is not preserved")
        if len(gens) == 0:    # handle the trivial group
            gens = [G.parent().identity_matrix()]
        from sage.libs.gap.libgap import libgap
        gap_gens = [libgap(matrix_gen) for matrix_gen in gens]
        gap_group = libgap.Group(gap_gens)
        FinitelyGeneratedMatrixGroup_gap.__init__(self,
                                                  degree,
                                                  base_ring,
                                                  gap_group,
                                                  category=category)

    def _repr_(self):
        r"""
        Return the string representation of this matrix group.

        OUTPUT: string

        EXAMPLES::

            sage: from sage.groups.matrix_gps.isometries import GroupOfIsometries
            sage: bil = Matrix(ZZ,2,[3,2,2,3])
            sage: gens = [-Matrix(ZZ,2,[0,1,1,0])]
            sage: O = GroupOfIsometries(2,ZZ,gens,bil)
            sage: O
            Group of isometries with 1 generator (
            [ 0 -1]
            [-1  0]
            )
        """
        n = self.ngens()
        from sage.repl.display.util import format_list
        if n > 5:
            return 'Group of isometries with %s generators ' % n
        if n == 1:
            return 'Group of isometries with %s generator %s' % (n, format_list(self.gens()))
        return 'Group of isometries with %s generators %s' % (n, format_list(self.gens()))

    def __reduce__(self):
        r"""
        Implement pickling.

        EXAMPLES::

            sage: from sage.groups.matrix_gps.isometries import GroupOfIsometries
            sage: bil = Matrix(ZZ,2,[3,2,2,3])
            sage: gens = [-Matrix(ZZ,2,[0,1,1,0])]
            sage: cat = Groups().Finite()
            sage: O = GroupOfIsometries(2, ZZ, gens, bil, category=cat)
            sage: loads(dumps(O)) == O
            True
        """
        args = (self.degree(), self.base_ring(),
                tuple(g.matrix() for g in self.gens()), self._invariant_bilinear_form,
                self.category(),
                False,
                self._invariant_submodule,
                self._invariant_quotient_module)
        return (GroupOfIsometries, args)

    def invariant_bilinear_form(self):
        r"""
        Return the symmetric bilinear form preserved by the orthogonal group.

        OUTPUT: the matrix defining the bilinear form

        EXAMPLES::

            sage: from sage.groups.matrix_gps.isometries import GroupOfIsometries
            sage: bil = Matrix(ZZ,2,[3,2,2,3])
            sage: gens = [-Matrix(ZZ,2,[0,1,1,0])]
            sage: O = GroupOfIsometries(2,ZZ,gens,bil)
            sage: O.invariant_bilinear_form()
            [3 2]
            [2 3]
        """
        return self._invariant_bilinear_form

    def _get_action_(self, S, op, self_on_left):
        """
        Provide the coercion system with an action.

        EXAMPLES::

            sage: from sage.groups.matrix_gps.isometries import GroupOfIsometries
            sage: bil = Matrix(ZZ,2,[3,2,2,3])
            sage: gens = [-Matrix(ZZ,2,[0,1,1,0])]
            sage: S = ZZ^2
            sage: T = S/(6*S)
            sage: O = GroupOfIsometries(2, ZZ, gens, bil, invariant_submodule=S, invariant_quotient_module=T)
            sage: O._get_action_(S, operator.mul, False)
            Right action by Group of isometries with 1 generator (
            [ 0 -1]
            [-1  0]
            ) on Ambient free module of rank 2 over the principal ideal domain Integer Ring
            sage: U = T.submodule([2*t for t in T.gens()])
            sage: u = U.an_element()
            sage: f = O.an_element()
            sage: u*f
            (0, 2)
        """
        import operator
        if op == operator.mul and not self_on_left:
            if S is self._invariant_submodule:
                return GroupActionOnSubmodule(self, S)
            if S is self._invariant_quotient_module:
                return GroupActionOnQuotientModule(self, S)
            from sage.modules.fg_pid.fgp_module import FGP_Module_class
            T = self._invariant_quotient_module
            if isinstance(S, FGP_Module_class):
                if S.is_submodule(T):
                    V = S.V()
                    if all(V == V * f.matrix() for f in self.gens()):
                        return GroupActionOnQuotientModule(self, S)
        return None

    def _check_matrix(self, x, *args):
        r"""
        Check whether the matrix ``x`` preserves the bilinear form.

        See :meth:`~sage.groups.matrix_gps.matrix_group._check_matrix`
        for details.

        EXAMPLES::

            sage: from sage.groups.matrix_gps.isometries import GroupOfIsometries
            sage: bil = Matrix(ZZ,2,[3,2,2,3])
            sage: gens = [-Matrix(ZZ,2,[0,1,1,0])]
            sage: O = GroupOfIsometries(2,ZZ,gens,bil)
            sage: g = matrix.identity(2)*2
            sage: O(g)
            Traceback (most recent call last):
            ...
            TypeError: matrix must be orthogonal with respect to the invariant form
        """
        F = self.invariant_bilinear_form()
        if x * F * x.transpose() != F:
            raise TypeError('matrix must be orthogonal '
                'with respect to the invariant form')


class GroupActionOnSubmodule(Action):
    r"""
    Matrix group action on a submodule from the right.

    INPUT:

    - ``MatrixGroup`` -- an instance of :class:`GroupOfIsometries`
    - ``submodule`` -- an invariant submodule
    - ``is_left`` -- boolean (default: ``False``)

    EXAMPLES::

        sage: from sage.groups.matrix_gps.isometries import GroupOfIsometries
        sage: S = span(ZZ, [[0,1]])
        sage: g = Matrix(QQ, 2, [1,0,0,-1])
        sage: G = GroupOfIsometries(2, ZZ, [g],
        ....:                       invariant_bilinear_form=matrix.identity(2),
        ....:                       invariant_submodule=S)
        sage: g = G.an_element()
        sage: x = S.an_element()
        sage: x*g
        (0, -1)
        sage: (x*g).parent()
        Free module of degree 2 and rank 1 over Integer Ring
        Echelon basis matrix:
        [0 1]
    """
    def __init__(self, MatrixGroup, submodule, is_left=False):
        r"""
        Initialize the action.

        TESTS::

            sage: from sage.groups.matrix_gps.isometries import GroupOfIsometries, GroupActionOnSubmodule
            sage: S = span(ZZ, [[0,1]])
            sage: g = Matrix(QQ, 2, [1,0,0,-1])
            sage: e = Matrix.identity(2)
            sage: G = GroupOfIsometries(2, ZZ, [g], e)
            sage: GroupActionOnSubmodule(G, S)
            Right action by Group of isometries with 1 generator (
            [ 1  0]
            [ 0 -1]
            ) on Free module of degree 2 and rank 1 over Integer Ring
            Echelon basis matrix:
            [0 1]
        """
        import operator
        Action.__init__(self, MatrixGroup, submodule, is_left, operator.mul)

    def _act_(self, g, a):
        r"""
        This defines the group action.

        INPUT:

        - ``g`` -- an element of the acting group

        - ``a`` -- an element of the invariant submodule

        OUTPUT: an element of the invariant submodule

        EXAMPLES::

            sage: from sage.groups.matrix_gps.isometries import GroupActionOnSubmodule
            sage: S = span(QQ,[[0,1]])
            sage: g = Matrix(QQ,2,[1,1,0,1/2])
            sage: G = MatrixGroup([g])
            sage: A = GroupActionOnSubmodule(G,S)
            sage: A
            Right action by Matrix group over Rational Field with 1 generators (
            [  1   1]
            [  0 1/2]
            ) on Vector space of degree 2 and dimension 1 over Rational Field
            Basis matrix:
            [0 1]
            sage: s = S.an_element()
            sage: g = G.an_element()
            sage: A(s,g).parent()
            Vector space of degree 2 and dimension 1 over Rational Field
            Basis matrix:
            [0 1]
        """
        if self.is_left():
            b = g.matrix() * a
        else:
            b = a * g.matrix()
        return a.parent()(b)


class GroupActionOnQuotientModule(Action):
    r"""
    Matrix group action on a quotient module from the right.

    INPUT:

    - ``MatrixGroup`` -- the group acting
      :class:`GroupOfIsometries`
    - ``submodule`` -- an invariant quotient module
    - ``is_left`` -- boolean (default: ``False``)

    EXAMPLES::

        sage: from sage.groups.matrix_gps.isometries import GroupOfIsometries
        sage: S = span(ZZ,[[0,1]])
        sage: Q = S/(6*S)
        sage: g = Matrix(QQ,2,[1,0,0,-1])
        sage: G = GroupOfIsometries(2, ZZ, [g], invariant_bilinear_form=matrix.identity(2), invariant_quotient_module=Q)
        sage: g = G.an_element()
        sage: x = Q.an_element()
        sage: x*g
        (5)
        sage: (x*g).parent()
        Finitely generated module V/W over Integer Ring with invariants (6)
    """
    def __init__(self, MatrixGroup, quotient_module, is_left=False):
        r"""
        Initialize the action.

        TESTS::

            sage: from sage.groups.matrix_gps.isometries import GroupOfIsometries
            sage: S = span(ZZ,[[0,1]])
            sage: Q = S/(6*S)
            sage: g = Matrix(QQ,2,[1,0,0,-1])
            sage: G = GroupOfIsometries(2, ZZ, [g], invariant_bilinear_form=matrix.identity(2), invariant_quotient_module=Q)
            sage: g = G.an_element()
            sage: x = Q.an_element()
            sage: x, x*g
            ((1), (5))
        """
        import operator
        Action.__init__(self, MatrixGroup, quotient_module, is_left, operator.mul)

    def _act_(self, g, a):
        r"""
        This defines the group action.

        INPUT:

        - ``g`` -- an element of the acting group

        - ``a`` -- an element of the invariant submodule

        OUTPUT: an element of the invariant quotient module

        EXAMPLES::

            sage: from sage.groups.matrix_gps.isometries import GroupActionOnQuotientModule
            sage: S = span(ZZ,[[0,1]])
            sage: Q = S/(6*S)
            sage: g = Matrix(QQ,2,[1,1,0,7])
            sage: G = MatrixGroup([g])
            sage: A = GroupActionOnQuotientModule(G,Q)
            sage: A
            Right action by Matrix group over Rational Field with 1 generators (
            [1 1]
            [0 7]
            ) on Finitely generated module V/W over Integer Ring with invariants (6)
            sage: q = Q.an_element()
            sage: g = G.an_element()
            sage: A(q,g).parent()
            Finitely generated module V/W over Integer Ring with invariants (6)
        """
        if self.is_left():
            b = g.matrix() * a.lift()
        else:
            b = a.lift() * g.matrix()
        return a.parent()(b)



# sage.doctest: needs sage.modules sage.rings.finite_rings
r"""
Guruswami-Sudan utility methods

AUTHORS:

- Johan S. R. Nielsen, original implementation (see [Nie]_ for details)
- David Lucas, ported the original implementation in Sage
"""

# ***************************************************************************
#       Copyright (C) 2015 David Lucas <david.lucas@inria.fr>
#                     2015 Johan S. R. Nielsen <jsrn@jsrn.dk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ***************************************************************************

from sage.arith.misc import integer_floor as floor
from sage.misc.functional import sqrt
from sage.rings.integer_ring import ZZ
from sage.rings.integer import Integer


def polynomial_to_list(p, len):
    r"""
    Return ``p`` as a list of its coefficients of length ``len``.

    INPUT:

    - ``p`` -- a polynomial

    - ``len`` -- integer; if ``len`` is smaller than the degree of ``p``, the
      returned list will be of size degree of ``p``, else it will be of size ``len``

    EXAMPLES::

        sage: from sage.coding.guruswami_sudan.utils import polynomial_to_list
        sage: F.<x> = GF(41)[]
        sage: p = 9*x^2 + 8*x + 37
        sage: polynomial_to_list(p, 4)
        [37, 8, 9, 0]
    """
    return list(p) + [0]*max(0, len-p.degree()-1)


def johnson_radius(n, d):
    r"""
    Return the Johnson-radius for the code length `n` and the minimum distance `d`.

    The Johnson radius is defined as `n - \sqrt(n(n-d))`.

    INPUT:

    - ``n`` -- integer; the length of the code
    - ``d`` -- integer; the minimum distance of the code

    EXAMPLES::

        sage: sage.coding.guruswami_sudan.utils.johnson_radius(250, 181)                # needs sage.symbolic
        -5*sqrt(690) + 250
    """
    return n - sqrt(n*(n-d))


def ligt(x):
    r"""
    Return the least integer greater than ``x``.

    EXAMPLES::

        sage: from sage.coding.guruswami_sudan.utils import ligt
        sage: ligt(41)
        42

    It works with any type of numbers (not only integers)::

        sage: ligt(41.041)
        42
    """
    return floor(x + 1)


def gilt(x):
    r"""
    Return the greatest integer smaller than ``x``.

    EXAMPLES::

        sage: from sage.coding.guruswami_sudan.utils import gilt
        sage: gilt(43)
        42

    It works with any type of numbers (not only integers)::

        sage: gilt(43.041)
        43
    """
    if x in ZZ:
        return Integer(x - 1)
    return floor(x)


def solve_degree2_to_integer_range(a, b, c):
    r"""
    Return the greatest integer range `[i_1, i_2]` such that
    `i_1 > x_1` and `i_2 < x_2` where `x_1, x_2` are the two zeroes of the
    equation in `x`: `ax^2+bx+c=0`.

    If there is no real solution to the equation, it returns an empty range
    with negative coefficients.

    INPUT:

    - ``a``, ``b`` and ``c`` -- coefficients of a second degree equation, ``a``
      being the coefficient of the higher degree term

    EXAMPLES::

        sage: from sage.coding.guruswami_sudan.utils import solve_degree2_to_integer_range
        sage: solve_degree2_to_integer_range(1, -5, 1)                                  # needs sage.symbolic
        (1, 4)

    If there is no real solution::

        sage: solve_degree2_to_integer_range(50, 5, 42)
        (-2, -1)
    """
    D = b**2 - 4*a*c
    if D < 0:
        return (-2,-1)
    sD = float(sqrt(D))
    minx, maxx = (-b-sD)/2.0/a , (-b+sD)/2.0/a
    mini, maxi = (ligt(minx), gilt(maxx))
    if mini > maxi:
        return (-2,-1)
    return (mini,maxi)


def _degree_of_vector(v, shifts=None):
    r"""
    Return the greatest degree among the entries of the polynomial vector `v`.

    INPUT:

    - ``v`` -- a vector of polynomials

    - ``shifts`` -- (default: ``None``) a list of integer shifts to consider
      ``v`` under, i.e. compute `\max(\deg v_i + s_i)`, where `s_1,\ldots, s_n`
      is the list of shifts.

      If ``None``, all shifts are considered as ``0``.

    EXAMPLES::

        sage: from sage.coding.guruswami_sudan.utils import _degree_of_vector
        sage: F.<x> = GF(7)[]
        sage: v = vector(F, [0, 1, x, x^2])
        sage: _degree_of_vector(v)
        2
        sage: _degree_of_vector(v, shifts=[10, 1, 0, -3])
        1
    """
    if not shifts:
        return max(vi.degree() for vi in v)
    if v.is_zero():
        return -1
    return max(degi + si for (degi, si) in zip([vi.degree() for vi in v ], shifts)
               if degi > -1)



# sage.doctest: needs sage.libs.pari sage.rings.number_field
"""
Local Representation Conditions
"""
########################################################################
# Class for keeping track of the local conditions for representability #
# of numbers by a quadratic form over ZZ (and eventually QQ also).     #
########################################################################
from __future__ import annotations
from copy import deepcopy

from sage.arith.misc import is_square, prime_divisors, valuation
from sage.misc.functional import numerator, denominator
from sage.quadratic_forms.extras import least_quadratic_nonresidue
from sage.rings.infinity import infinity
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ


class QuadraticFormLocalRepresentationConditions:
    """
    A class for dealing with the local conditions of a
    quadratic form, and checking local representability of numbers.

    EXAMPLES::

        sage: Q4 = DiagonalQuadraticForm(ZZ, [1,1,1,1])
        sage: Q4.local_representation_conditions()
        This form represents the p-adic integers Z_p for all primes p except
        [].  For these and the reals, we have:
             Reals:   [0, +Infinity]
        sage: Q4.is_locally_represented_number(1)
        True
        sage: Q4.is_locally_universal_at_all_primes()
        True
        sage: Q4.is_locally_universal_at_all_places()
        False
        sage: L = [m  for m in range(-5, 100)  if Q4.is_locally_represented_number(m)]
        sage: L == list(range(100))
        True

    ::

        sage: Q3 = DiagonalQuadraticForm(ZZ, [1,1,1])
        sage: Q3.local_representation_conditions()
        This form represents the p-adic integers Z_p for all primes p except
        [2].  For these and the reals, we have:
             Reals:   [0, +Infinity]
             p = 2:   [0, 0, 0, +Infinity, 0, 0, 0, 0]
        sage: E = [m  for m in range(100)  if not Q3.is_locally_represented_number(m)]
        sage: E1 = [m  for m in range(1,100)  if m / 2**(2 * (valuation(m,2) // 2)) % 8 == 7]
        sage: E == E1
        True
        sage: E
        [7, 15, 23, 28, 31, 39, 47, 55, 60, 63, 71, 79, 87, 92, 95]

    ::

        sage: Q2 = DiagonalQuadraticForm(ZZ, [1,1])
        sage: Q2.local_representation_conditions()
        This 2-dimensional form represents the p-adic integers of even
        valuation for all primes p except [2].
        For these and the reals, we have:
             Reals:   [0, +Infinity]
             p = 2:   [0, +Infinity, 0, +Infinity, 0, +Infinity, 0, +Infinity]
        sage: Q2.is_locally_universal_at_all_places()
        False
        sage: Q2.is_locally_universal_at_all_primes()
        False
        sage: L = [m  for m in range(-5, 25)  if Q2.is_locally_represented_number(m)]
        sage: L1 = [0] + [m for m in range(1, 25)
        ....:             if len([p for p in prime_factors(squarefree_part(ZZ(m)))
        ....:                       if (p % 4) == 3]) % 2 == 0]
        sage: L == L1
        True
        sage: L
        [0, 1, 2, 4, 5, 8, 9, 10, 13, 16, 17, 18, 20, 21]

    ::

        sage: Q1 = DiagonalQuadraticForm(ZZ, [1])
        sage: Q1.local_representation_conditions()
        This 1-dimensional form only represents square multiples of 1.
        sage: L = [m  for m in range(100)  if Q1.is_locally_represented_number(m)]
        sage: L
        [0, 1, 4, 9, 16, 25, 36, 49, 64, 81]

    ::

        sage: Q0 = DiagonalQuadraticForm(ZZ, [])
        sage: Q0.local_representation_conditions()
        This 0-dimensional form only represents zero.
        sage: L = [m  for m in range(100)  if Q0.is_locally_represented_number(m)]
        sage: L
        [0]
    """
    def __init__(self, Q):
        r"""
        Take a :class:`QuadraticForm` and computes its local conditions (if
        they do not already exist).  The ``recompute_flag`` overrides the
        previously computed conditions if they exist, and stores the
        new conditions.

        INPUT:

        - ``Q`` -- Quadratic form over `\ZZ`

        OUTPUT: a :class:`QuadraticFormLocalRepresentationConditions` object

        EXAMPLES::

            sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1])
            sage: from sage.quadratic_forms.quadratic_form__local_representation_conditions import QuadraticFormLocalRepresentationConditions
            sage: QuadraticFormLocalRepresentationConditions(Q)
            This form represents the p-adic integers Z_p for all primes p except
            [].  For these and the reals, we have:
                 Reals:   [0, +Infinity]
        """
        # Check that the form Q is integer-valued (we can relax this later)
        if Q.base_ring() != ZZ:
            raise TypeError("We require that the quadratic form be defined over ZZ (integer-values) for now.")

        # Basic structure initialization
        self.local_repn_array = []    # List of all local conditions
        self.dim = Q.dim()       # We allow this to be any nonnegative integer.
        self.exceptional_primes = [infinity]

        # Deal with the special cases of 0 and 1-dimensional forms
        if self.dim == 0:
            self.coeff = None
            return
        if self.dim == 1:
            self.coeff = Q[0, 0]
            return

        self.coeff = None

        # Compute the local conditions at the real numbers (i.e. "p = infinity")
        # ----------------------------------------------------------------------
        M = Q.matrix()
        E = M.eigenspaces_left()
        M_eigenvalues = [E[i][0] for i in range(len(E))]

        pos_flag = infinity
        neg_flag = infinity

        for e in M_eigenvalues:
            if e > 0:
                pos_flag = 0
            elif e < 0:
                neg_flag = 0

        real_vec = [infinity, pos_flag, neg_flag, None, None, None, None, None, None]
        self.local_repn_array.append(real_vec)

        # Compute the local conditions for representability:
        # --------------------------------------------------
        N = Q.level()
        level_primes = prime_divisors(N)

        # Make a table of local normal forms for each p | N
        local_normal_forms = [Q.local_normal_form(p) for p in level_primes]

        # Check local representability conditions for each prime
        for i, p in enumerate(level_primes):
            tmp_local_repn_vec = [p, None, None, None, None, None, None, None, None]
            sqclass = self.squareclass_vector(p)

            # Check the representability in each Z_p squareclass
            for j, m in enumerate(sqclass):
                k = 0
                repn_flag = False

                while ((not repn_flag) and (m < 4 * N * p * p)):
                    if local_normal_forms[i].local_density(p, m) > 0:
                        tmp_local_repn_vec[j + 1] = k
                        repn_flag = True
                    k = k + 1
                    m = m * p * p

                # If we're not represented, write "infinity" to signify
                # that this squareclass is fully obstructed
                if not repn_flag:
                    tmp_local_repn_vec[j + 1] = infinity

            # Test if the conditions at p give exactly Z_p when dim >=3, or
            # if we represent the elements of even valuation >= 2 when dim = 2.
            omit_flag = True
            if self.dim >= 2:
                # Check that all entries are zero or 'None'
                for x in tmp_local_repn_vec[1:]:
                    if not ((x == 0) or (x is None)):
                        omit_flag = False

            # Add the results for this prime if there is a congruence obstruction
            if not omit_flag:
                self.local_repn_array.append(tmp_local_repn_vec)
                self.exceptional_primes.append(p)

    def __repr__(self) -> str:
        r"""
        Print the local conditions.

        OUTPUT: string

        .. TODO::

            Improve the output for the real numbers, and special output for locally universality.
            Also give names to the squareclasses, so it's clear what the output means! =)

        EXAMPLES::

            sage: Q = DiagonalQuadraticForm(ZZ, [1,1])
            sage: from sage.quadratic_forms.quadratic_form__local_representation_conditions import QuadraticFormLocalRepresentationConditions
            sage: C = QuadraticFormLocalRepresentationConditions(Q)
            sage: C.__repr__()
            'This 2-dimensional form represents the p-adic integers of even\nvaluation for all primes p except [2].\nFor these and the reals, we have:\n     Reals:   [0, +Infinity]\n     p = 2:   [0, +Infinity, 0, +Infinity, 0, +Infinity, 0, +Infinity]\n'
        """
        if self.dim == 0:
            out_str = "This 0-dimensional form only represents zero."
        elif self.dim == 1:
            out_str = "This 1-dimensional form only represents square multiples of " + str(self.coeff) + "."
        elif self.dim == 2:
            out_str = "This 2-dimensional form represents the p-adic integers of even\n"
            out_str += "valuation for all primes p except " + str(self.exceptional_primes[1:]) + ".\n"
            out_str += "For these and the reals, we have:\n"
        else:
            out_str = "This form represents the p-adic integers Z_p for all primes p except \n"
            out_str += str(self.exceptional_primes[1:]) + ".  For these and the reals, we have:\n"

        for v in self.local_repn_array:
            if v[0] == infinity:
                out_str += "     " + "Reals:   " + str(v[1:3]) + "\n"
            elif v[0] == 2:
                out_str += "     " + "p = 2:   " + str(v[1:]) + "\n"
            else:
                out_str += "     " + "p = " + str(v[0]) + ":   " + str(v[1:5]) + "\n"

        return out_str

    def __eq__(self, right) -> bool:
        """
        Determine if two sets of local conditions are equal.

        INPUT:

        - ``right`` -- a QuadraticFormLocalRepresentationConditions object

        OUTPUT: boolean

        EXAMPLES::

             sage: Q1 = DiagonalQuadraticForm(ZZ, [1,1])
             sage: Q2 = DiagonalQuadraticForm(ZZ, [1,1,1])
             sage: Q3 = DiagonalQuadraticForm(ZZ, [1,3,5,7])
             sage: Q4 = DiagonalQuadraticForm(ZZ, [1,1,1,1])

             sage: Q1.local_representation_conditions() == Q2.local_representation_conditions()
             False
             sage: Q1.local_representation_conditions() == Q3.local_representation_conditions()
             False
             sage: Q1.local_representation_conditions() == Q4.local_representation_conditions()
             False
             sage: Q2.local_representation_conditions() == Q3.local_representation_conditions()
             False
             sage: Q3.local_representation_conditions() == Q4.local_representation_conditions()
             True
        """
        if not isinstance(right, QuadraticFormLocalRepresentationConditions):
            return False

        # Check the dimensions agree when they affect the kind of representation conditions.
        if ((self.dim <= 2) or (right.dim <= 2)) and self.dim != right.dim:
            return False

        # Check equality by dimension
        if self.dim == 0:
            return True
        if self.dim == 1:
            return self.coeff == right.coeff     # Compare coefficients in dimension 1 (since ZZ has only one unit square)
        return ((self.exceptional_primes == right.exceptional_primes)
                and (self.local_repn_array == right.local_repn_array))

    def squareclass_vector(self, p) -> list:
        """
        Return a list of integers which are normalized
        representatives for the `p`-adic rational squareclasses
        (or the real squareclasses) at the prime `p`.

        INPUT:

        - ``p`` -- a positive prime number or "infinity"

        OUTPUT: list of integers

        EXAMPLES::

            sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
            sage: from sage.quadratic_forms.quadratic_form__local_representation_conditions import QuadraticFormLocalRepresentationConditions
            sage: C = QuadraticFormLocalRepresentationConditions(Q)
            sage: C.squareclass_vector(5)
            [1, 2, 5, 10]
        """
        if p == infinity:
            return [1, -1]
        if p == 2:
            return [1, 3, 5, 7, 2, 6, 10, 14]
        r = least_quadratic_nonresidue(p)
        return [1, r, p, p * r]

    def local_conditions_vector_for_prime(self, p) -> list:
        """
        Return a local representation vector for the (possibly infinite) prime `p`.

        INPUT:

        - ``p`` -- a positive prime number.  (Is 'infinity' allowed here?)

        OUTPUT: list of integers

        EXAMPLES::

            sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
            sage: from sage.quadratic_forms.quadratic_form__local_representation_conditions import QuadraticFormLocalRepresentationConditions
            sage: C = QuadraticFormLocalRepresentationConditions(Q)
            sage: C.local_conditions_vector_for_prime(2)
            [2, 0, 0, 0, +Infinity, 0, 0, 0, 0]
            sage: C.local_conditions_vector_for_prime(3)
            [3, 0, 0, 0, 0, None, None, None, None]
        """
        # Check if p is non-generic
        if p in self.exceptional_primes:
            return deepcopy(self.local_repn_array[self.exceptional_primes.index(p)])

        # Otherwise, generate a vector at this (finite) prime
        if self.dim >= 3:
            if p == 2:
                return [2, 0, 0, 0, 0, 0, 0, 0, 0]
            return [p, 0, 0, 0, 0, None, None, None, None]

        if self.dim == 2:
            if p == 2:
                return [2, 0, 0, 0, 0, infinity, infinity, infinity, infinity]
            return [p, 0, 0, infinity, infinity, None, None, None, None]

        if self.dim == 1:
            v = [p, None, None, None, None, None, None, None, None]
            sqclass = self.squareclass_vector(p)

            for i, sqi in enumerate(sqclass):
                if QQ(self.coeff / sqi).is_padic_square(p):    # Note:This should happen only once!
                    nu = valuation(self.coeff / sqi, p) / 2    # UNUSED VARIABLE !
                else:
                    v[i + 1] = infinity

        elif self.dim == 0:
            if p == 2:
                return [2, infinity, infinity, infinity, infinity, infinity, infinity, infinity, infinity]
            return [p, infinity, infinity, infinity, infinity, None, None, None, None]

        raise RuntimeError("the stored dimension should be a nonnegative integer")

    def is_universal_at_prime(self, p) -> bool:
        r"""
        Determine if the (integer-valued/rational) quadratic form represents all of `\ZZ_p`.

        INPUT:

        - ``p`` -- a positive prime number or "infinity"

        OUTPUT: boolean

        EXAMPLES::

            sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
            sage: from sage.quadratic_forms.quadratic_form__local_representation_conditions import QuadraticFormLocalRepresentationConditions
            sage: C = QuadraticFormLocalRepresentationConditions(Q)
            sage: C.is_universal_at_prime(2)
            False
            sage: C.is_universal_at_prime(3)
            True
            sage: C.is_universal_at_prime(infinity)
            False
        """
        # Check if the prime behaves generically for n >= 3.
        if self.dim >= 3 and p not in self.exceptional_primes:
            return True

        # Check if the prime behaves generically for n <= 2.
        if self.dim <= 2 and p not in self.exceptional_primes:
            return False

        # Check if the prime is "infinity" (for the reals)
        if p == infinity:
            v = self.local_repn_array[0]
            if p != v[0]:
                raise RuntimeError("Error... The first vector should be for the real numbers!")
            return (v[1:3] == [0, 0])     # True iff the form is indefinite

        # Check non-generic "finite" primes
        v = self.local_conditions_vector_for_prime(p)
        Zp_univ_flag = True
        for nu in v[1:]:
            if (nu is not None) and ((nu != 0) or (nu == infinity)):
                Zp_univ_flag = False
        return Zp_univ_flag

    def is_universal_at_all_finite_primes(self) -> bool:
        r"""
        Determine if the quadratic form represents `\ZZ_p` for all finite/non-archimedean primes.

        OUTPUT: boolean

        EXAMPLES::

            sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
            sage: from sage.quadratic_forms.quadratic_form__local_representation_conditions import QuadraticFormLocalRepresentationConditions
            sage: C = QuadraticFormLocalRepresentationConditions(Q)
            sage: C.is_universal_at_all_finite_primes()
            False

        ::

            sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1])
            sage: from sage.quadratic_forms.quadratic_form__local_representation_conditions import QuadraticFormLocalRepresentationConditions
            sage: C = QuadraticFormLocalRepresentationConditions(Q)
            sage: C.is_universal_at_all_finite_primes()
            True
        """
        # Check if dim <= 2.
        if self.dim <= 2:
            return False

        # Check that all non-generic finite primes are universal
        # Omit p = "infinity" here
        return all(self.is_universal_at_prime(p)
                   for p in self.exceptional_primes[1:])

    def is_universal_at_all_places(self) -> bool:
        r"""
        Determine if the quadratic form represents `\ZZ_p` for all
        finite/non-archimedean primes, and represents all real numbers.

        OUTPUT: boolean

        EXAMPLES::

            sage: from sage.quadratic_forms.quadratic_form__local_representation_conditions import QuadraticFormLocalRepresentationConditions

        ::

            sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
            sage: C = QuadraticFormLocalRepresentationConditions(Q)
            sage: C.is_universal_at_all_places()
            False

        ::

            sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1])
            sage: C = QuadraticFormLocalRepresentationConditions(Q)
            sage: C.is_universal_at_all_places()
            False

        ::

            sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1,-1])
            sage: C = QuadraticFormLocalRepresentationConditions(Q)     # long time (8.5 s)
            sage: C.is_universal_at_all_places()                        # long time
            True
        """
        # Check if dim <= 2.
        if self.dim <= 2:
            return False

        # Check that all non-generic finite primes are universal
        return all(self.is_universal_at_prime(p)
                   for p in self.exceptional_primes)

    def is_locally_represented_at_place(self, m, p) -> bool:
        """
        Determine if the rational number `m` is locally represented by the
        quadratic form at the (possibly infinite) prime `p`.

        INPUT:

        - ``m`` -- integer

        - ``p`` -- a positive prime number or "infinity"

        OUTPUT: boolean

        EXAMPLES::

            sage: from sage.quadratic_forms.quadratic_form__local_representation_conditions import QuadraticFormLocalRepresentationConditions

            sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
            sage: C = QuadraticFormLocalRepresentationConditions(Q)
            sage: C.is_locally_represented_at_place(7, 2)
            False
            sage: C.is_locally_represented_at_place(1, 3)
            True
            sage: C.is_locally_represented_at_place(-1, infinity)
            False
            sage: C.is_locally_represented_at_place(1, infinity)
            True
            sage: C.is_locally_represented_at_place(0, infinity)
            True
        """
        # Sanity Check
        if m not in QQ:
            raise TypeError(f"m = {m} is not a rational number")

        # Representing zero
        if m == 0:
            return True

        # 0-dim'l forms
        if self.dim == 0:   # Here m != 0
            return False

        # 1-dim'l forms
        if self.dim == 1:
            m1 = QQ(m) / self.coeff
            if p == infinity:
                return m1 > 0
            return (valuation(m1, p) >= 0) and m1.is_padic_square(p)

        # >= 2-dim'l forms
        local_vec = self.local_conditions_vector_for_prime(p)

        # Check the real place
        if p == infinity:
            if m > 0:
                return local_vec[1] == 0
            if m < 0:
                return local_vec[2] == 0
            # m == 0
            return True

        # Check at a finite place
        sqclass = self.squareclass_vector(p)
        for s in sqclass:
            if (QQ(m) / s).is_padic_square(p):
                nu = valuation(m // s, p)
                return local_vec[sqclass.index(s) + 1] <= (nu / 2)

    def is_locally_represented(self, m) -> bool:
        r"""
        Determine if the rational number `m` is locally represented by
        the quadratic form (allowing vectors with coefficients in `\ZZ_p` at all
        places).

        INPUT:

        - ``m`` -- integer

        OUTPUT: boolean

        EXAMPLES::

            sage: from sage.quadratic_forms.quadratic_form__local_representation_conditions import QuadraticFormLocalRepresentationConditions

            sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
            sage: C = QuadraticFormLocalRepresentationConditions(Q)
            sage: C.is_locally_represented(7)
            False
            sage: C.is_locally_represented(28)
            False
            sage: C.is_locally_represented(11)
            True
            sage: C.is_locally_represented(QQ(1)/QQ(2))
            False
        """
        # Representing zero
        if m == 0:
            return True

        # 0-dim'l forms
        if self.dim == 0:    # Here m != 0
            return False

        # 1-dim'l forms
        if self.dim == 1:
            m1 = m / self.coeff
            return (m1 in ZZ) and is_square(m1)

        # Check the generic primes (when n = 2 or n >= 3)
        m_primes = prime_divisors(numerator(m) * denominator(m))
        for p in m_primes:
            if p not in self.exceptional_primes:
                val = valuation(m, p)
                if val < 0:
                    return False

        # Check the non-generic primes (when n = 2 or n >= 3)
        for p in self.exceptional_primes:
            if not self.is_locally_represented_at_place(m, p):
                return False

        # If we got here, we're locally represented!
        return True

# ---  End of QuadraticFormLocalRepresentationConditions Class ---


def local_representation_conditions(self, recompute_flag=False, silent_flag=False):
    r"""
    .. WARNING::

        This only works correctly for forms in >=3 variables,
        which are locally universal at almost all primes!

    This class finds the local conditions for a number to be integrally
    represented by an integer-valued quadratic form.  These conditions
    are stored in ``self.__local_representability_conditions`` and
    consist of a list of 9 element vectors, with one for each prime
    with a local obstruction (though only the first 5 are meaningful
    unless `p=2`).  The first element is always the prime `p` where the
    local obstruction occurs, and the next 8 (or 4) entries represent
    square-classes in the `p`-adic integers `\ZZ_p`, and are labeled by the
    `\QQ_p` square-classes `t\cdot (\QQ_p)^2` with `t` given as follows:

    - for `p > 2`, ``[ *  1  u  p  u p  *  *  *  * ]``,

    - for `p = 2`, ``[ *  1  3  5  7  2  6  10  14 ]``.

    The integer appearing in each place tells us how `p`-divisible a
    number needs to be in that square-class in order to be locally
    represented by `Q`.  A negative number indicates that the entire `\QQ_p`
    square-class is not represented, while a positive number `x` indicates
    that `t\cdot p^{(2\cdot x)} (\ZZ_p)^2` is locally represented but `t\cdot p^{(2\cdot (x-1))}`
    `(\ZZ_p)^2` is not.

    As an example, the vector ``[2  3  0  0  0  0  2  0  infinity]``
    tells us that all positive integers are locally represented at `p=2`
    except those of the forms:

    - `2^6\cdot u\cdot r^2` with `u = 1` (mod 8)

    - `2^5\cdot u\cdot r^2` with `u = 3` (mod 8)

    - `2\cdot u\cdot r^2` with `u = 7` (mod 8)

    At the real numbers, the vector which looks like ``[infinity, 0, infinity, None, None, None, None, None, None]``
    means that `Q` is negative definite (i.e., the 0 tells us all
    positive reals are represented).  The real vector always appears,
    and is listed before the other ones.

    OUTPUT:

    A list of 9-element vectors describing the representation
    obstructions at primes dividing the level.

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [])
        sage: Q.local_representation_conditions()
        This 0-dimensional form only represents zero.

        sage: Q = DiagonalQuadraticForm(ZZ, [5])
        sage: Q.local_representation_conditions()
        This 1-dimensional form only represents square multiples of 5.

        sage: Q1 = DiagonalQuadraticForm(ZZ, [1,1])
        sage: Q1.local_representation_conditions()
        This 2-dimensional form represents the p-adic integers of even
        valuation for all primes p except [2].
        For these and the reals, we have:
         Reals:   [0, +Infinity]
         p = 2:   [0, +Infinity, 0, +Infinity, 0, +Infinity, 0, +Infinity]

        sage: Q1 = DiagonalQuadraticForm(ZZ, [1,1,1])
        sage: Q1.local_representation_conditions()
        This form represents the p-adic integers Z_p for all primes p except
        [2].  For these and the reals, we have:
         Reals:   [0, +Infinity]
         p = 2:   [0, 0, 0, +Infinity, 0, 0, 0, 0]

        sage: Q1 = DiagonalQuadraticForm(ZZ, [1,1,1,1])
        sage: Q1.local_representation_conditions()
        This form represents the p-adic integers Z_p for all primes p except
        [].  For these and the reals, we have:
         Reals:   [0, +Infinity]

        sage: Q1 = DiagonalQuadraticForm(ZZ, [1,3,3,3])
        sage: Q1.local_representation_conditions()
        This form represents the p-adic integers Z_p for all primes p except
        [3].  For these and the reals, we have:
         Reals:   [0, +Infinity]
         p = 3:   [0, 1, 0, 0]

        sage: Q2 = DiagonalQuadraticForm(ZZ, [2,3,3,3])
        sage: Q2.local_representation_conditions()
        This form represents the p-adic integers Z_p for all primes p except
        [3].  For these and the reals, we have:
         Reals:   [0, +Infinity]
         p = 3:   [1, 0, 0, 0]

        sage: Q3 = DiagonalQuadraticForm(ZZ, [1,3,5,7])
        sage: Q3.local_representation_conditions()
        This form represents the p-adic integers Z_p for all primes p except
        [].  For these and the reals, we have:
         Reals:   [0, +Infinity]
    """
    # Recompute the local conditions if they do not exist or the recompute_flag is set.
    if not hasattr(self, "__local_representability_conditions") or recompute_flag:
        self.__local_representability_conditions = QuadraticFormLocalRepresentationConditions(self)

    # Return the local conditions if the silent_flag is not set.
    if not silent_flag:
        return self.__local_representability_conditions


def is_locally_universal_at_prime(self, p) -> bool:
    r"""
    Determine if the (integer-valued/rational) quadratic form represents all of `\ZZ_p`.

    INPUT:

    - ``p`` -- a positive prime number or "infinity"

    OUTPUT: boolean

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,5,7])
        sage: Q.is_locally_universal_at_prime(2)
        True
        sage: Q.is_locally_universal_at_prime(3)
        True
        sage: Q.is_locally_universal_at_prime(5)
        True
        sage: Q.is_locally_universal_at_prime(infinity)
        False

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
        sage: Q.is_locally_universal_at_prime(2)
        False
        sage: Q.is_locally_universal_at_prime(3)
        True
        sage: Q.is_locally_universal_at_prime(5)
        True
        sage: Q.is_locally_universal_at_prime(infinity)
        False

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,-1])
        sage: Q.is_locally_universal_at_prime(infinity)
        True
    """
    self.local_representation_conditions(silent_flag=True)
    return self.__local_representability_conditions.is_universal_at_prime(p)


def is_locally_universal_at_all_primes(self) -> bool:
    r"""
    Determine if the quadratic form represents `\ZZ_p` for all finite/non-archimedean primes.

    OUTPUT: boolean

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,5,7])
        sage: Q.is_locally_universal_at_all_primes()
        True

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1])
        sage: Q.is_locally_universal_at_all_primes()
        True

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
        sage: Q.is_locally_universal_at_all_primes()
        False
    """
    self.local_representation_conditions(silent_flag=True)
    return self.__local_representability_conditions.is_universal_at_all_finite_primes()


def is_locally_universal_at_all_places(self) -> bool:
    r"""
    Determine if the quadratic form represents `\ZZ_p` for all
    finite/non-archimedean primes, and represents all real numbers.

    OUTPUT: boolean

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,5,7])
        sage: Q.is_locally_universal_at_all_places()
        False

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1])
        sage: Q.is_locally_universal_at_all_places()
        False

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1,-1])
        sage: Q.is_locally_universal_at_all_places()        # long time (8.5 s)
        True
    """
    self.local_representation_conditions(silent_flag=True)
    return self.__local_representability_conditions.is_universal_at_all_places()


def is_locally_represented_number_at_place(self, m, p) -> bool:
    """
    Determine if the rational number `m` is locally represented by the
    quadratic form at the (possibly infinite) prime `p`.

    INPUT:

    - ``m`` -- integer

    - ``p`` -- a prime number > 0 or 'infinity'

    OUTPUT: boolean

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
        sage: Q.is_locally_represented_number_at_place(7, infinity)
        True
        sage: Q.is_locally_represented_number_at_place(7, 2)
        False
        sage: Q.is_locally_represented_number_at_place(7, 3)
        True
        sage: Q.is_locally_represented_number_at_place(7, 5)
        True
        sage: Q.is_locally_represented_number_at_place(-1, infinity)
        False
        sage: Q.is_locally_represented_number_at_place(-1, 2)
        False

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1,-1])
        sage: Q.is_locally_represented_number_at_place(7, infinity)     # long time (8.5 s)
        True
        sage: Q.is_locally_represented_number_at_place(7, 2)            # long time
        True
        sage: Q.is_locally_represented_number_at_place(7, 3)            # long time
        True
        sage: Q.is_locally_represented_number_at_place(7, 5)            # long time
        True
    """
    self.local_representation_conditions(silent_flag=True)
    return self.__local_representability_conditions.is_locally_represented_at_place(m, p)


def is_locally_represented_number(self, m) -> bool:
    """
    Determine if the rational number `m` is locally represented
    by the quadratic form.

    INPUT:

    - ``m`` -- integer

    OUTPUT: boolean

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
        sage: Q.is_locally_represented_number(2)
        True
        sage: Q.is_locally_represented_number(7)
        False
        sage: Q.is_locally_represented_number(-1)
        False
        sage: Q.is_locally_represented_number(28)
        False
        sage: Q.is_locally_represented_number(0)
        True
    """
    self.local_representation_conditions(silent_flag=True)
    return self.__local_representability_conditions.is_locally_represented(m)



# sage.doctest: needs sage.combinat sage.modules
r"""
Fully packed loops

AUTHORS:

- Vincent Knight, James Campbell, Kevin Dilks, Emily Gunawan (2015): Initial version
- Vincent Delecroix (2017): cleaning and enhanced plotting function
"""
# ****************************************************************************
#       Copyright (C) 2015 Vincent Knight <vincent.knight@gmail.com>
#                          James Campbell <james.campbell@tanti.org.uk>
#                          Kevin Dilks <kdilks@gmail.com>
#                          Emily Gunawan <egunawan@umn.edu>
#                     2017 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful, but
#    WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.inherit_comparison import InheritComparisonClasscallMetaclass
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.parent import Parent
from sage.structure.element import parent, Element

from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.combinat.six_vertex_model import (SquareIceModel,
                                            SixVertexConfiguration,
                                            SixVertexModel)
from sage.combinat.alternating_sign_matrix import AlternatingSignMatrix

from sage.misc.decorators import options
from sage.matrix.constructor import matrix
from sage.arith.misc import factorial
from sage.rings.integer import Integer
from sage.misc.misc_c import prod

# edges of a fpl in terms of the six vertex possible configurations
R = (1, 0)
L = (-1, 0)
U = (0, 1)
D = (0, -1)

FPL_edges = (
    # 0 UD    1 RD,   2 UR,   3 LR,   4 LD    5 LU
    ((D, U), (L, D), (D, R), (R, L), (L, U), (R, U)),  # even
    ((R, L), (R, U), (L, U), (D, U), (D, R), (L, D))   # odd
)

FPL_turns = (
    # 0 UD          1 RD          2 UR          3 LR          4 LD          5 LU
    ({U: U, D: D}, {R: D, U: L}, {U: R, L: D}, {L: L, R: R}, {R: U, D: L}, {L: U, D: R}),  # even
    ({L: L, R: R}, {L: U, D: R}, {R: U, D: L}, {U: U, D: D}, {U: R, L: D}, {R: D, U: L})   # odd
)


def _make_color_list(n, colors=None, color_map=None, randomize=False):
    r"""
    TESTS::

        sage: import numpy as np
        sage: if int(np.version.short_version[0]) > 1:
        ....:     _ = np.set_printoptions(legacy="1.25")
        sage: from sage.combinat.fully_packed_loop import _make_color_list
        sage: _make_color_list(5)
        sage: _make_color_list(5, ['blue', 'red'])
        ['blue', 'red', 'blue', 'red', 'blue']
        sage: _make_color_list(5, color_map='summer')
        [(0.0, 0.5, 0.4),
         (0.25098039215686274, 0.6254901960784314, 0.4),
         (0.5019607843137255, 0.7509803921568627, 0.4),
         (0.7529411764705882, 0.8764705882352941, 0.4),
         (1.0, 1.0, 0.4)]
        sage: l = _make_color_list(8, ['blue', 'red'], randomize=True)
        sage: len(l)
        8
        sage: l.count('blue')
        4
        sage: l.count('red')
        4
    """
    if colors:
        dim = len(colors)
        colors = [colors[i % dim] for i in range(n)]

    elif color_map:
        from matplotlib import cm
        if color_map not in cm.datad:
            raise ValueError('unknown color map %s' % color_map)
        cmap = cm.__dict__[color_map]
        colors = [cmap(i / float(n - 1))[:3] for i in range(n)]

    if colors and randomize:
        from sage.misc.prandom import shuffle
        shuffle(colors)

    return colors


class FullyPackedLoop(Element, metaclass=InheritComparisonClasscallMetaclass):
    r"""
    A class for fully packed loops.

    A fully packed loop is a collection of non-intersecting lattice paths on a square
    grid such that every vertex is part of some path, and the paths are either closed
    internal loops or have endpoints corresponding to alternate points on the
    boundary [Pro2001]_. They are known to be in bijection with alternating sign
    matrices.

    .. SEEALSO::

        :class:`AlternatingSignMatrix`

    To each fully packed loop, we assign a link pattern, which is the non-crossing
    matching attained by seeing which points on the boundary are connected
    by open paths in the fully packed loop.

    We can create a fully packed loop using the corresponding alternating sign
    matrix and also extract the link pattern::

        sage: A = AlternatingSignMatrix([[0, 0, 1], [0, 1, 0], [1, 0, 0]])
        sage: fpl = FullyPackedLoop(A)
        sage: fpl.link_pattern()
        [(1, 4), (2, 3), (5, 6)]
        sage: fpl
           │         │
           │         │
           + ── +    +
                │    │
                │    │
        ── +    +    + ──
           │    │
           │    │
           +    + ── +
           │         │
           │         │
        sage: B = AlternatingSignMatrix([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        sage: fplb = FullyPackedLoop(B)
        sage: fplb.link_pattern()
        [(1, 6), (2, 5), (3, 4)]
        sage: fplb
           │         │
           │         │
           +    + ── +
           │    │
           │    │
        ── +    +    + ──
                │    │
                │    │
           + ── +    +
           │         │
           │         │

    The class also has a plot method::

        sage: fpl.plot()                                                                # needs sage.plot
        Graphics object consisting of 3 graphics primitives

    which gives:

    .. PLOT::
        :width: 200 px

        A = AlternatingSignMatrix([[0, 0, 1], [0, 1, 0], [1, 0, 0]])
        fpl = FullyPackedLoop(A)
        p = fpl.plot()
        sphinx_plot(p)

    Note that we can also create a fully packed loop from a six vertex model configuration::

        sage: S = SixVertexModel(3, boundary_conditions='ice').from_alternating_sign_matrix(A)
        sage: S
            ^    ^    ^
            |    |    |
        --> # -> # -> # <--
            ^    ^    |
            |    |    V
        --> # -> # <- # <--
            ^    |    |
            |    V    V
        --> # <- # <- # <--
            |    |    |
            V    V    V
        sage: fpl = FullyPackedLoop(S)
        sage: fpl
           │         │
           │         │
           + ── +    +
                │    │
                │    │
        ── +    +    + ──
           │    │
           │    │
           +    + ── +
           │         │
           │         │

    Once we have a fully packed loop we can obtain the corresponding alternating sign matrix::

        sage: fpl.to_alternating_sign_matrix()
        [0 0 1]
        [0 1 0]
        [1 0 0]

    Here are some more examples using bigger ASMs::

        sage: A = AlternatingSignMatrix([[0,1,0,0],[0,0,1,0],[1,-1,0,1],[0,1,0,0]])
        sage: S = SixVertexModel(4, boundary_conditions='ice').from_alternating_sign_matrix(A)
        sage: fpl = FullyPackedLoop(S)
        sage: fpl.link_pattern()
        [(1, 2), (3, 6), (4, 5), (7, 8)]
        sage: fpl
           │         │
           │         │
           + ── + ── +    + ──
                          │
                          │
        ── +    + ── + ── +
           │    │
           │    │
           +    +    + ── + ──
           │    │    │
           │    │    │
        ── +    +    + ── +
                │         │
                │         │

        sage: m = AlternatingSignMatrix([[0,0,1,0,0,0],
        ....:                            [1,0,-1,0,1,0],
        ....:                            [0,0,0,1,0,0],
        ....:                            [0,1,0,0,-1,1],
        ....:                            [0,0,0,0,1,0],
        ....:                            [0,0,1,0,0,0]])
        sage: fpl = FullyPackedLoop(m)
        sage: fpl.link_pattern()
        [(1, 12), (2, 7), (3, 4), (5, 6), (8, 9), (10, 11)]
        sage: fpl
           │         │         │
           │         │         │
           + ── +    +    + ── +    + ──
                │    │    │         │
                │    │    │         │
        ── + ── +    +    + ── + ── +
                     │
                     │
           + ── +    + ── + ── +    + ──
           │    │              │    │
           │    │              │    │
        ── +    +    + ── +    +    +
                │    │    │    │    │
                │    │    │    │    │
           + ── +    + ── +    +    + ──
           │                   │
           │                   │
        ── +    + ── + ── +    + ── +
                │         │         │
                │         │         │

        sage: m = AlternatingSignMatrix([[0,1,0,0,0,0,0],
        ....:                            [1,-1,0,0,1,0,0],
        ....:                            [0,0,0,1,0,0,0],
        ....:                            [0,1,0,0,-1,1,0],
        ....:                            [0,0,0,0,1,0,0],
        ....:                            [0,0,1,0,-1,0,1],
        ....:                            [0,0,0,0,1,0,0]])
        sage: fpl = FullyPackedLoop(m)
        sage: fpl.link_pattern()
        [(1, 2), (3, 4), (5, 6), (7, 8), (9, 14), (10, 11), (12, 13)]
        sage: fpl
           │         │         │         │
           │         │         │         │
           + ── + ── +    + ── +    + ── +
                          │         │
                          │         │
        ── + ── + ── +    + ── + ── +    + ──
                     │                   │
                     │                   │
           + ── +    + ── + ── +    + ── +
           │    │              │    │
           │    │              │    │
        ── +    +    + ── +    +    +    + ──
                │    │    │    │    │    │
                │    │    │    │    │    │
           + ── +    + ── +    +    + ── +
           │                   │
           │                   │
        ── +    + ── + ── +    +    + ── + ──
                │         │    │    │
                │         │    │    │
           + ── +    + ── +    +    + ── +
           │         │         │         │
           │         │         │         │

    Gyration on an alternating sign matrix/fully packed loop ``fpl``
    of the link pattern corresponding to ``fpl``::

        sage: ASMs = AlternatingSignMatrices(3).list()
        sage: ncp = FullyPackedLoop(ASMs[1]).link_pattern() # fpl's gyration orbit size is 2
        sage: rotated_ncp=[]
        sage: for a, b in ncp:
        ....:     for i in range(5):
        ....:         a,b=a%6+1,b%6+1;
        ....:     rotated_ncp.append((a,b))
        sage: PerfectMatching(ASMs[1].gyration().to_fully_packed_loop().link_pattern()) ==\
        ....:     PerfectMatching(rotated_ncp)
        True

        sage: fpl = FullyPackedLoop(ASMs[0])
        sage: ncp = fpl.link_pattern() # fpl's gyration size is 3
        sage: rotated_ncp=[]
        sage: for a, b in ncp:
        ....:     for i in range(5):
        ....:         a,b=a%6+1,b%6+1;
        ....:     rotated_ncp.append((a,b))
        sage: PerfectMatching(ASMs[0].gyration().to_fully_packed_loop().link_pattern()) ==\
        ....:     PerfectMatching(rotated_ncp)
        True

        sage: mat = AlternatingSignMatrix([[0,0,1,0,0,0,0],[1,0,-1,0,1,0,0],
        ....:     [0,0,1,0,0,0,0],[0,1,-1,0,0,1,0],[0,0,1,0,0,0,0],[0,0,0,1,0,0,0],[0,0,0,0,0,0,1]])
        sage: fpl = FullyPackedLoop(mat) # n=7
        sage: ncp = fpl.link_pattern()
        sage: rotated_ncp=[]
        sage: for a, b in ncp:
        ....:     for i in range(13):
        ....:         a,b=a%14+1,b%14+1;
        ....:     rotated_ncp.append((a,b))
        sage: PerfectMatching(mat.gyration().to_fully_packed_loop().link_pattern()) ==\
        ....:     PerfectMatching(rotated_ncp)
        True

        sage: mat = AlternatingSignMatrix([[0,0,0,1,0,0], [0,0,1,-1,1,0],
        ....:     [0,1,0,0,-1,1], [1,0,-1,1,0,0], [0,0,1,0,0,0], [0,0,0,0,1,0]])
        sage: fpl = FullyPackedLoop(mat) # n =6
        sage: ncp = fpl.link_pattern()
        sage: rotated_ncp=[]
        sage: for a, b in ncp:
        ....:     for i in range(11):
        ....:         a,b=a%12+1,b%12+1;
        ....:     rotated_ncp.append((a,b))
        sage: PerfectMatching(mat.gyration().to_fully_packed_loop().link_pattern()) ==\
        ....:     PerfectMatching(rotated_ncp)
        True

    More examples:

    We can initiate a fully packed loop using an alternating sign matrix::

        sage: A = AlternatingSignMatrix([[0, 0, 1], [0, 1, 0], [1, 0, 0]])
        sage: fpl = FullyPackedLoop(A)
        sage: fpl
            │         │
            │         │
            + ── +    +
                 │    │
                 │    │
         ── +    +    + ──
            │    │
            │    │
            +    + ── +
            │         │
            │         │
        sage: FullyPackedLoops(3)(A) == fpl
        True

    We can also input a matrix::

        sage: FullyPackedLoop([[0, 0, 1], [0, 1, 0], [1, 0, 0]])
            │         │
            │         │
            + ── +    +
                 │    │
                 │    │
         ── +    +    + ──
            │    │
            │    │
            +    + ── +
            │         │
            │         │
        sage: FullyPackedLoop([[0, 0, 1], [0, 1, 0], [1, 0, 0]]) ==\
        ....: FullyPackedLoops(3)([[0, 0, 1], [0, 1, 0], [1, 0, 0]])
        True

    Otherwise we initiate a fully packed loop using a six vertex model::

        sage: S = SixVertexModel(3, boundary_conditions='ice').from_alternating_sign_matrix(A)
        sage: fpl = FullyPackedLoop(S)
        sage: fpl
            │         │
            │         │
            + ── +    +
                 │    │
                 │    │
         ── +    +    + ──
            │    │
            │    │
            +    + ── +
            │         │
            │         │

        sage: FullyPackedLoops(3)(S) == FullyPackedLoop(S)
        True

        sage: fpl.six_vertex_model().to_alternating_sign_matrix()
        [0 0 1]
        [0 1 0]
        [1 0 0]

    We can also input the matrix associated to a six vertex model::

        sage: SixVertexModel(2)([[3,1],[5,3]])
            ^    ^
            |    |
        --> # <- # <--
            |    ^
            V    |
        --> # -> # <--
            |    |
            V    V

        sage: FullyPackedLoop([[3,1],[5,3]])
            │
            │
            +    + ──
            │    │
            │    │
         ── +    +
                 │
                 │

        sage: FullyPackedLoops(2)([[3,1],[5,3]]) == FullyPackedLoop([[3,1],[5,3]])
        True

    Note that the matrix corresponding to a six vertex model without
    the ice boundary condition is not allowed::

        sage: SixVertexModel(2)([[3,1],[5,5]])
            ^    ^
            |    |
        --> # <- # <--
            |    ^
            V    V
        --> # -> # -->
            |    |
            V    V

        sage: FullyPackedLoop([[3,1],[5,5]])
        Traceback (most recent call last):
        ...
        ValueError: invalid alternating sign matrix

        sage: FullyPackedLoops(2)([[3,1],[5,5]])
        Traceback (most recent call last):
        ...
        ValueError: invalid alternating sign matrix

    Note that if anything else is used to generate the fully packed loop an error will occur::

        sage: fpl = FullyPackedLoop(5)
        Traceback (most recent call last):
        ...
        ValueError: invalid alternating sign matrix

        sage: fpl = FullyPackedLoop((1, 2, 3))
        Traceback (most recent call last):
        ...
        ValueError: the alternating sign matrices must be square

        sage: SVM = SixVertexModel(3)[0]
        sage: FullyPackedLoop(SVM)
        Traceback (most recent call last):
        ...
        ValueError: invalid alternating sign matrix

    REFERENCES:

    - [Pro2001]_
    - [Str2015]_
    """
    @staticmethod
    def __classcall_private__(cls, generator):
        """
        Create a FPL.

        EXAMPLES::

            sage: A = AlternatingSignMatrix([[1, 0, 0],[0, 1, 0],[0, 0, 1]])
            sage: FullyPackedLoop(A)
               │         │
               │         │
               +    + ── +
               │    │
               │    │
            ── +    +    + ──
                    │    │
                    │    │
               + ── +    +
               │         │
               │         │

            sage: SVM = SixVertexModel(4, boundary_conditions='ice')[0]
            sage: FullyPackedLoop(SVM)
               │         │
               │         │
               +    + ── +    + ──
               │    │         │
               │    │         │
            ── +    +    + ── +
                    │    │
                    │    │
               + ── +    +    + ──
               │         │    │
               │         │    │
            ── +    + ── +    +
                    │         │
                    │         │
        """
        if isinstance(generator, AlternatingSignMatrix):
            SVM = generator.to_six_vertex_model()
        elif isinstance(generator, SquareIceModel.Element):
            SVM = generator
        elif isinstance(generator, SixVertexConfiguration):
            # Check that this is an ice square model
            generator = SixVertexModel(generator.parent()._nrows,
                                       boundary_conditions='ice')(generator)
            M = generator.to_alternating_sign_matrix().to_matrix()
            AlternatingSignMatrix(M)
            SVM = generator
        else:  # Not ASM nor SVM
            try:
                SVM = AlternatingSignMatrix(generator).to_six_vertex_model()
            except (TypeError, ValueError):
                generator = matrix(generator)
                generator = SixVertexModel(generator.nrows(), boundary_conditions='ice')(generator)
                # Check that this is an ice square model
                generator.to_alternating_sign_matrix()
                SVM = generator

        if not SVM:
            raise TypeError('generator for FullyPackedLoop must either be an '
                            'AlternatingSignMatrix or a SquareIceModel.Element')
        FPLs = FullyPackedLoops(len(SVM))
        return FPLs(generator)

    def __init__(self, parent, generator):
        """
        Initialise object, can take ASM of FPL as generator.

        TESTS::

            sage: A = AlternatingSignMatrix([[0, 0, 1], [0, 1, 0], [1, 0, 0]])
            sage: fpl = FullyPackedLoop(A)
            sage: TestSuite(fpl).run()
        """
        if isinstance(generator, AlternatingSignMatrix):
            self._six_vertex_model = generator.to_six_vertex_model()
        elif isinstance(generator, SquareIceModel.Element):
            self._six_vertex_model = generator

        Element.__init__(self, parent)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: A = AlternatingSignMatrix([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
            sage: fpl = FullyPackedLoop(A)
            sage: fpl
               │         │
               │         │
               +    + ── +
               │    │
               │    │
            ── +    +    + ──
                    │    │
                    │    │
               + ── +    +
               │         │
               │         │

            sage: A = AlternatingSignMatrix([[0,1,0,0],[0,0,1,0],[1,-1,0,1],[0,1,0,0]])
            sage: S = SixVertexModel(4, boundary_conditions='ice').from_alternating_sign_matrix(A)
            sage: fpl = FullyPackedLoop(S)
            sage: fpl
               │         │
               │         │
               + ── + ── +    + ──
                              │
                              │
            ── +    + ── + ── +
               │    │
               │    │
               +    +    + ── + ──
               │    │    │
               │    │    │
            ── +    +    + ── +
                    │         │
                    │         │
        """
        # List are in the order of URDL
        # One set of rules for how to draw around even vertex, one set of rules for odd vertex
        n = len(self._six_vertex_model) - 1
        ascii1 = [[r'     ', ' ─', r'     ', '─ '],  # LR
                 [r'  │  ', '  ', r'     ', '─ '],  # LU
                 [r'     ', '  ', r'  │  ', '─ '],  # LD
                 [r'  │  ', '  ', r'  │  ', '  '],  # UD
                 [r'  │  ', ' ─', r'     ', '  '],  # UR
                 [r'     ', ' ─', r'  │  ', '  ']]  # RD

        ascii2 = [[r'  │  ', '  ', r'  │  ', '  '],  # LR
                 [r'     ', ' ─', r'  │  ', '  '],  # LU
                 [r'  │  ', ' ─', r'     ', '  '],  # LD
                 [r'     ', ' ─', r'     ', '─ '],  # UD
                 [r'     ', '  ', r'  │  ', '─ '],  # UR
                 [r'  │  ', '  ', r'     ', '─ ']]  # RD
        ret = '  '
        # Do the top line
        for i, entry in enumerate(self._six_vertex_model[0]):
            ret += '     ' if i % 2 else '  │  '

        plus_sign = '+'

        # Do the meat of the ascii art
        for j, row in enumerate(self._six_vertex_model):
            ret += '\n  '
            # Do the top row
            for i, entry in enumerate(row):
                if not (i + j) % 2:
                    ret += ascii1[entry][0]
                else:
                    ret += ascii2[entry][0]
            ret += '\n'

            # Do the left-most entry
            if not j % 2:
                ret += '  '
            else:
                ret += ' ─'

            # Do the middle row
            for i, entry in enumerate(row):
                if (i + j) % 2 == 0:
                    ret += ascii1[entry][3] + plus_sign + ascii1[entry][1]
                else:
                    ret += ascii2[entry][3] + plus_sign + ascii2[entry][1]

            # Do the right-most entry
            if not (j + n) % 2:
                ret += '  '
            else:
                ret += '─ '

            # Do the bottom row
            ret += '\n  '
            for i, entry in enumerate(row):
                if not (i + j) % 2:
                    ret += ascii1[entry][2]
                else:
                    ret += ascii2[entry][2]

        # Do the bottom line
        ret += '\n  '
        for i, entry in enumerate(self._six_vertex_model[-1]):
            ret += '  │  ' if (i + n + 1) % 2 else '     '

        return ret

    def _richcmp_(self, other, op):
        """
        Check equality or inequality.

        EXAMPLES::

            sage: A = AlternatingSignMatrices(3)
            sage: M = A.random_element()
            sage: FullyPackedLoop(M) == M.to_fully_packed_loop()
            True

            sage: FullyPackedLoop(A([[1, 0, 0],[0, 1, 0],[0, 0, 1]])) ==\
            ....:    FullyPackedLoop(A([[1, 0, 0],[0, 0, 1],[0, 1, 0]]))
            False

            sage: FullyPackedLoop(M) == M
            False

            sage: M = A.random_element()
            sage: FullyPackedLoop(M) != M.to_fully_packed_loop()
            False

            sage: f0 = FullyPackedLoop(A([[1, 0, 0],[0, 1, 0],[0, 0, 1]]))
            sage: f1 = FullyPackedLoop(A([[1, 0, 0],[0, 0, 1],[0, 1, 0]]))
            sage: f0 != f1
            True
        """
        return self._six_vertex_model._richcmp_(other._six_vertex_model, op)

    def to_alternating_sign_matrix(self):
        """
        Return the alternating sign matrix corresponding to this class.

        .. SEEALSO::

            :class:`AlternatingSignMatrix`

        EXAMPLES::

            sage: A = AlternatingSignMatrix([[0, 1, 0], [1, -1, 1], [0, 1, 0]])
            sage: S = SixVertexModel(3, boundary_conditions='ice').from_alternating_sign_matrix(A)
            sage: fpl = FullyPackedLoop(S)
            sage: fpl.to_alternating_sign_matrix()
            [ 0  1  0]
            [ 1 -1  1]
            [ 0  1  0]
            sage: A = AlternatingSignMatrix([[0,1,0,0],[0,0,1,0],[1,-1,0,1],[0,1,0,0]])
            sage: S = SixVertexModel(4, boundary_conditions='ice').from_alternating_sign_matrix(A)
            sage: fpl = FullyPackedLoop(S)
            sage: fpl.to_alternating_sign_matrix()
            [ 0  1  0  0]
            [ 0  0  1  0]
            [ 1 -1  0  1]
            [ 0  1  0  0]
        """
        return self._six_vertex_model.to_alternating_sign_matrix()

    @options(link=True, loop=True, loop_fill=False)
    def plot(self, **options):
        r"""
        Return a graphical object of the Fully Packed Loop.

        Each option can be specified separately for links (the curves that join
        boundary points) and the loops. In order to do so, you need to prefix
        its name with either ``'link_'`` or ``'loop_'``. As an example, setting
        ``color='red'`` will color both links and loops in red while setting
        ``link_color='red'`` will only apply the color option for the links.

        INPUT:

        - ``link``, ``loop`` -- boolean (default: ``True``); whether to plot the links
          or the loops

        - ``color``, ``link_color``, ``loop_color`` -- (optional) string or RGB triple

        - ``colors``, ``link_colors``, ``loop_colors`` -- (optional) list of colors

        - ``color_map``, ``link_color_map``, ``loop_color_map`` -- (string,
          optional) a name of a matplotlib color map for the link or the loop

        - ``link_color_randomize`` -- boolean (default: ``False``); when
          ``link_colors`` or ``link_color_map`` is specified it randomizes
          its order. Setting this option to ``True`` makes it unlikely to
          have two neighboring links with the same color.

        - ``loop_fill`` -- boolean (default: ``False``); whether to fill the interior of the loops

        EXAMPLES:

        To plot the fully packed loop associated to the following alternating sign
        matrix

        .. MATH::

            \begin{pmatrix} 0&1&1 \\ 1&-1&1 \\ 0&1&0 \end{pmatrix}

        simply do::

            sage: A = AlternatingSignMatrix([[0, 1, 0], [1, -1, 1], [0, 1, 0]])
            sage: fpl = FullyPackedLoop(A)
            sage: fpl.plot()                                                            # needs sage.plot
            Graphics object consisting of 3 graphics primitives

        The resulting graphics is as follows

        .. PLOT::
            :width: 200 px

            A = AlternatingSignMatrix([[0, 1, 0], [1, -1, 1], [0, 1, 0]])
            fpl = FullyPackedLoop(A)
            p = fpl.plot()
            sphinx_plot(p)

        You can also have the three links in different colors with::

            sage: A = AlternatingSignMatrix([[0, 1, 0], [1, -1, 1], [0, 1, 0]])
            sage: fpl = FullyPackedLoop(A)
            sage: fpl.plot(link_color_map='rainbow')                                    # needs sage.plot
            Graphics object consisting of 3 graphics primitives

        .. PLOT::
            :width: 200 px

            A = AlternatingSignMatrix([[0, 1, 0], [1, -1, 1], [0, 1, 0]])
            fpl = FullyPackedLoop(A)
            p = fpl.plot(link_color_map='rainbow')
            sphinx_plot(p)

        You can plot the 42 fully packed loops of size `4 \times 4` using::

            sage: G = [fpl.plot(link_color_map='winter', loop_color='black')            # needs sage.plot
            ....:      for fpl in FullyPackedLoops(4)]
            sage: graphics_array(G, 7, 6)                                               # needs sage.plot
            Graphics Array of size 7 x 6

        .. PLOT::
            :width: 600 px

            G = [fpl.plot(link_color_map='winter', loop_color='black') for fpl in FullyPackedLoops(4)]
            p = graphics_array(G, 7, 6)
            sphinx_plot(p)

        Here is an example of a `20 \times 20` fully packed loop::

            sage: s = "00000000000+0000000000000000+00-0+00000000000+00-00+0-+00000\
            ....: 0000+-00+00-+00000000+00-0000+0000-+00000000+000-0+0-+0-+000\
            ....: 000+-000+-00+0000000+-+-000+00-+0-000+000+-000+-0+0000000-0+\
            ....: 0000+0-+0-+00000-+00000+-+0-0+-00+0000000000+-0000+0-00+0000\
            ....: 000000+0-000+000000000000000+0000-00+00000000+0000-000+00000\
            ....: 00+0-00+0000000000000000+-0000+000000-+000000+00-0000+-00+00\
            ....: 00000000+-0000+00000000000000+0000000000"
            sage: a = matrix(20, [{'0':0, '+':1, '-': -1}[i] for i in s])
            sage: fpl = FullyPackedLoop(a)
            sage: fpl.plot(loop_fill=True, loop_color_map='rainbow')                    # needs sage.plot
            Graphics object consisting of 27 graphics primitives

        .. PLOT::
            :width: 400 px

            s = "00000000000+0000000000000000+00-0+00000000000+00-00+0-+00000\
            0000+-00+00-+00000000+00-0000+0000-+00000000+000-0+0-+0-+000\
            000+-000+-00+0000000+-+-000+00-+0-000+000+-000+-0+0000000-0+\
            0000+0-+0-+00000-+00000+-+0-0+-00+0000000000+-0000+0-00+0000\
            000000+0-000+000000000000000+0000-00+00000000+0000-000+00000\
            00+0-00+0000000000000000+-0000+000000-+000000+00-0000+-00+00\
            00000000+-0000+00000000000000+0000000000"
            a = matrix(20, [{'0':0, '+':1, '-': -1}[i] for i in s])
            p = FullyPackedLoop(a).plot(loop_fill=True, loop_color_map='rainbow')
            sphinx_plot(p)
        """
        from sage.plot.graphics import Graphics
        from sage.plot.line import line2d
        from sage.plot.polygon import polygon2d

        link_options = {}
        loop_options = {}
        for k, v in options.items():
            if k == 'link':
                link = v
            elif k == 'loop':
                loop = v
            elif k.startswith('link_'):
                link_options[k[5:]] = v
            elif k.startswith('loop_'):
                loop_options[k[5:]] = v
            else:
                link_options[k] = v
                loop_options[k] = v

        sv = self._six_vertex_model
        n = len(sv)

        # LR boundaries => odd sum
        # UD boundaries => even sum
        rank = self.parent()._boundary_index
        unrank = self.parent()._boundary
        seen = [False] * (2*n)

        squares = set((i, j) for i in range(n) for j in range(n))

        colors = _make_color_list(2*n,
                colors=link_options.pop('colors', None),
                color_map=link_options.pop('color_map', None),
                randomize=link_options.pop('color_randomize', False))

        G = Graphics()
        for i in range(2*n):
            if seen[i]:
                continue
            orbit = self._link_or_loop_from(unrank(i))
            j = rank(orbit[-1])
            seen[i] = seen[j] = True
            squares.difference_update(orbit)

            if link:
                if colors:
                    link_options['color'] = colors.pop()

                # make it upside down
                orbit = [(j, n - i - 1) for i, j in orbit]
                G += line2d(orbit, **link_options)

        loops = []
        while squares:
            orbit = self._link_or_loop_from(squares.pop())
            loops.append(orbit)
            squares.difference_update(orbit)

        if loop:
            colors = _make_color_list(len(loops),
                    colors=loop_options.pop('colors', None),
                    color_map=loop_options.pop('color_map', None),
                    randomize=loop_options.pop('color_randomize', False))

            fill = loop_options.pop('fill')

            for orbit in loops:
                if colors:
                    loop_options['color'] = colors.pop()

                # make it upside down
                orbit = [(j, n - i - 1) for i, j in orbit]

                if fill:
                    G += polygon2d(orbit, **loop_options)
                else:
                    G += line2d(orbit, **loop_options)

        G.axes(False)
        G.set_aspect_ratio(1)
        return G

    def gyration(self):
        r"""
        Return the fully packed loop obtained by applying gyration
        to the alternating sign matrix in bijection with ``self``.

        Gyration was first defined in [Wie2000]_ as an action on
        fully-packed loops.

        EXAMPLES::

            sage: A = AlternatingSignMatrix([[1, 0, 0],[0, 1, 0],[0, 0, 1]])
            sage: fpl = FullyPackedLoop(A)
            sage: fpl.gyration().to_alternating_sign_matrix()
            [0 0 1]
            [0 1 0]
            [1 0 0]
            sage: asm = AlternatingSignMatrix([[0, 0, 1],[1, 0, 0],[0, 1, 0]])
            sage: f = FullyPackedLoop(asm)
            sage: f.gyration().to_alternating_sign_matrix()
            [0 1 0]
            [0 0 1]
            [1 0 0]
        """
        return FullyPackedLoop(self.to_alternating_sign_matrix().gyration())

    def _link_or_loop_from(self, pos, d0=None):
        r"""
        Return the coordinates of the line passing through ``pos``.

        EXAMPLES:

        A link::

            sage: fpl = FullyPackedLoops(4).first()
            sage: fpl._link_or_loop_from((2,2))
            [(0, 4), (0, 3), (1, 3), (1, 2), (2, 2), (3, 2), (3, 1), (4, 1)]
            sage: fpl._link_or_loop_from((-1, 0))
            [(-1, 0), (0, 0), (1, 0), (1, -1)]

        A loop::

            sage: a = AlternatingSignMatrix([[0,1,0,0], [0,0,0,1], [1,0,0,0], [0,0,1,0]])
            sage: fpl = FullyPackedLoop(a)
            sage: fpl._link_or_loop_from((1,1))
            [(1, 1), (2, 1), (2, 2), (1, 2), (1, 1)]
        """
        orbit = [pos]
        sv = self._six_vertex_model
        n = len(sv)
        i, j = pos

        # deal with boundary cases
        if i < -1 or i > n or j < -1 or j > n:
            raise ValueError('indices out of range')
        if (i == -1 or i == n) and not (i + j) % 2:
            raise ValueError('left and right boundary values must have odd sum')
        if (j == -1 or j == n) and (i + j) % 2:
            raise ValueError('up and down boundary values must have even sum')

        if i == -1:
            d = R
        elif i == n:
            d = L
        elif j == -1:
            d = U
        elif j == n:
            d = D
        elif d0 is None:
            d = FPL_edges[(i + j) % 2][sv[i][j]][0]
        elif d0 in FPL_edges[(i + j) % 2][sv[i][j]]:
            d = d0
        else:
            raise ValueError('invalid direction')

        # compute the link or loop
        while True:
            i += d[0]
            j += d[1]
            orbit.append((i, j))
            if (i, j) == orbit[0] or i == -1 or j == -1 or i == n or j == n:
                break

            conf = sv[i][j]
            parity = (i + j) % 2

            d = FPL_turns[parity][conf][d]
            if d is None:
                raise RuntimeError

        if i == -1 or j == -1 or i == n or j == n:
            i0, j0 = orbit[0]
            if d0 is None and i0 != -1 and i0 != n and j0 != -1 and j0 != n:
                # only half of a link -> compute the other half
                i1, j1 = orbit[1]
                d = (i0-i1, j0-j1)
                orbit2 = self._link_or_loop_from(orbit[1], d)
                assert orbit2[0] == (i1, j1) and orbit2[1] == (i0, j0)
                return orbit2[:1:-1] + orbit
            return orbit
        return orbit

    def link_pattern(self):
        r"""
        Return a link pattern corresponding to a fully packed loop.

        Here we define a link pattern `LP` to be a partition of the list
        `[1, ..., 2k]` into 2-element sets (such a partition is also known as
        a perfect matching) such that the following non-crossing condition holds:
        Let the numbers `1, ..., 2k` be written on the perimeter of a circle.
        For every 2-element set `(a,b)` of the partition `LP`, draw an arc
        linking the two numbers `a` and `b`. We say that `LP` is non-crossing
        if every arc can be drawn so that no two arcs intersect.

        Since every endpoint of a fully packed loop `fpl` is connected to a different
        endpoint, there is a natural surjection from the fully packed loops on an
        nxn grid onto the link patterns on the list `[1, \dots, 2n]`.
        The pairs of connected endpoints of a fully packed loop `fpl` correspond to
        the 2-element tuples of the corresponding link pattern.

        .. SEEALSO::

            :class:`PerfectMatching`

        .. NOTE::

            by convention, we choose the top left vertex to be even.
            See [Pro2001]_ and [Str2015]_.

        EXAMPLES:

        We can extract the underlying link pattern (a non-crossing
        partition) from a fully packed loop::

            sage: A = AlternatingSignMatrix([[0, 1, 0], [1, -1, 1], [0, 1, 0]])
            sage: fpl = FullyPackedLoop(A)
            sage: fpl.link_pattern()
            [(1, 2), (3, 6), (4, 5)]

            sage: B = AlternatingSignMatrix([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
            sage: fpl = FullyPackedLoop(B)
            sage: fpl.link_pattern()
            [(1, 6), (2, 5), (3, 4)]

        Gyration on an alternating sign matrix/fully packed loop ``fpl``
        corresponds to a rotation (i.e. a becomes a-1 mod 2n)
        of the link pattern corresponding to ``fpl``::

            sage: ASMs = AlternatingSignMatrices(3).list()
            sage: ncp = FullyPackedLoop(ASMs[1]).link_pattern()
            sage: rotated_ncp=[]
            sage: for a, b in ncp:
            ....:     for i in range(5):
            ....:         a,b=a%6+1,b%6+1;
            ....:     rotated_ncp.append((a,b))
            sage: PerfectMatching(ASMs[1].gyration().to_fully_packed_loop().link_pattern()) ==\
            ....:     PerfectMatching(rotated_ncp)
            True

            sage: fpl = FullyPackedLoop(ASMs[0])
            sage: ncp = fpl.link_pattern()
            sage: rotated_ncp=[]
            sage: for a, b in ncp:
            ....:     for i in range(5):
            ....:         a,b=a%6+1,b%6+1;
            ....:     rotated_ncp.append((a,b))
            sage: PerfectMatching(ASMs[0].gyration().to_fully_packed_loop().link_pattern()) ==\
            ....:     PerfectMatching(rotated_ncp)
            True

            sage: mat = AlternatingSignMatrix([[0,0,1,0,0,0,0],[1,0,-1,0,1,0,0],[0,0,1,0,0,0,0],
            ....:     [0,1,-1,0,0,1,0],[0,0,1,0,0,0,0],[0,0,0,1,0,0,0],[0,0,0,0,0,0,1]])
            sage: fpl = FullyPackedLoop(mat) # n=7
            sage: ncp = fpl.link_pattern()
            sage: rotated_ncp=[]
            sage: for a, b in ncp:
            ....:     for i in range(13):
            ....:         a,b=a%14+1,b%14+1;
            ....:     rotated_ncp.append((a,b))
            sage: PerfectMatching(mat.gyration().to_fully_packed_loop().link_pattern()) ==\
            ....:     PerfectMatching(rotated_ncp)
            True

            sage: mat = AlternatingSignMatrix([[0,0,0,1,0,0], [0,0,1,-1,1,0], [0,1,0,0,-1,1], [1,0,-1,1,0,0],
            ....:     [0,0,1,0,0,0], [0,0,0,0,1,0]])
            sage: fpl = FullyPackedLoop(mat)
            sage: ncp = fpl.link_pattern()
            sage: rotated_ncp=[]
            sage: for a, b in ncp:
            ....:     for i in range(11):
            ....:         a,b=a%12+1,b%12+1;
            ....:     rotated_ncp.append((a,b))
            sage: PerfectMatching(mat.gyration().to_fully_packed_loop().link_pattern()) ==\
            ....:     PerfectMatching(rotated_ncp)
            True

        TESTS:

        We test previous two bugs which showed up when this method is called twice::

            sage: A = AlternatingSignMatrices(6)
            sage: B = A.random_element()
            sage: C = FullyPackedLoop(B)
            sage: D = C.link_pattern()
            sage: E = C.link_pattern()
            sage: D == E
            True
        """
        link_pattern = []
        n = len(self._six_vertex_model)
        seen = [False] * (2*n)
        unrank = self.parent()._boundary
        rank = self.parent()._boundary_index
        sv = self._six_vertex_model

        for k in range(2*n):
            if seen[k]:
                continue

            i, j = unrank(k)

            # initial direction
            if i == -1:
                d = R
            elif i == n:
                d = L
            elif j == -1:
                d = U
            elif j == n:
                d = D

            # go through the link
            while True:
                i += d[0]
                j += d[1]
                if i == -1 or j == -1 or i == n or j == n:
                    break

                conf = sv[i][j]
                parity = (i + j) % 2
                d = FPL_turns[parity][conf][d]

            # update seen and link_pattern
            l = rank((i, j))
            seen[k] = seen[l] = True
            link_pattern.append((k+1, l+1))

        return link_pattern

    def six_vertex_model(self):
        """
        Return the underlying six vertex model configuration.

        EXAMPLES::

            sage: B = AlternatingSignMatrix([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
            sage: fpl = FullyPackedLoop(B)
            sage: fpl
               │         │
               │         │
               +    + ── +
               │    │
               │    │
            ── +    +    + ──
                    │    │
                    │    │
               + ── +    +
               │         │
               │         │
            sage: fpl.six_vertex_model()
                ^    ^    ^
                |    |    |
            --> # <- # <- # <--
                |    ^    ^
                V    |    |
            --> # -> # <- # <--
                |    |    ^
                V    V    |
            --> # -> # -> # <--
                |    |    |
                V    V    V
        """
        return self._six_vertex_model


class FullyPackedLoops(Parent, UniqueRepresentation):
    r"""
    Class of all fully packed loops on an  `n \times n` grid.

    They are known to be in bijection with alternating sign matrices.

    .. SEEALSO::

        :class:`AlternatingSignMatrices`

    INPUT:

    - ``n`` -- the number of row (and column) or grid

    EXAMPLES:

    This will create an instance to manipulate the fully packed loops of size 3::

        sage: FPLs = FullyPackedLoops(3)
        sage: FPLs
        Fully packed loops on a 3x3 grid
        sage: FPLs.cardinality()
        7

    When using the square ice model, it is known that the number of
    configurations is equal to the number of alternating sign matrices::

        sage: M = FullyPackedLoops(1)
        sage: len(M)
        1
        sage: M = FullyPackedLoops(4)
        sage: len(M)
        42
        sage: all(len(SixVertexModel(n, boundary_conditions='ice'))
        ....:     == FullyPackedLoops(n).cardinality() for n in range(1, 7))
        True
    """

    def __init__(self, n):
        r"""
        Initialize ``self``.

        TESTS::

            sage: FPLs = FullyPackedLoops(3)
            sage: TestSuite(FPLs).run()
        """
        self._n = n
        Parent.__init__(self, category=FiniteEnumeratedSets())

    def __iter__(self):
        """
        Iterate through ``self``.

        EXAMPLES::

            sage: FPLs = FullyPackedLoops(2)
            sage: len(FPLs)
            2
        """
        for X in SixVertexModel(self._n, boundary_conditions='ice'):
            yield self.element_class(self, X)

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        TESTS::

            sage: FPLs = FullyPackedLoops(4); FPLs
            Fully packed loops on a 4x4 grid
        """
        return "Fully packed loops on a %sx%s grid" % (self._n, self._n)

    def __contains__(self, fpl):
        """
        Check if ``fpl`` is in ``self``.

        TESTS::

            sage: FPLs = FullyPackedLoops(3)
            sage: FullyPackedLoop(AlternatingSignMatrix([[0,1,0],[1,0,0],[0,0,1]])) in FPLs
            True
            sage: FullyPackedLoop(AlternatingSignMatrix([[0,1,0],[1,-1,1],[0,1,0]])) in FPLs
            True
            sage: FullyPackedLoop(AlternatingSignMatrix([[0, 1],[1,0]])) in FPLs
            False
            sage: FullyPackedLoop(AlternatingSignMatrix([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1]])) in FPLs
            False
            sage: [1,2,3] in FPLs
            False
        """
        return parent(fpl) is self

    def _element_constructor_(self, generator):
        """
        Construct an element of ``self``.

        EXAMPLES::

            sage: FPLs = FullyPackedLoops(4)
            sage: M = [[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1]]
            sage: A = AlternatingSignMatrix(M)
            sage: elt = FullyPackedLoop(A)
            sage: FPL = FPLs(elt); FPL
               │         │
               │         │
               +    + ── +    + ──
               │    │         │
               │    │         │
            ── +    +    + ── +
                    │    │
                    │    │
               + ── +    +    + ──
               │         │    │
               │         │    │
            ── +    + ── +    +
                    │         │
                    │         │

            sage: FPLs(A) == FPL
            True

            sage: FPLs(M) == FPL
            True

            sage: FPLs(FPL._six_vertex_model) == FPL
            True

            sage: FPL.parent() is FPLs
            True

            sage: FPL = FullyPackedLoops(2)
            sage: FPL([[3,1],[5,3]])
               │
               │
               +    + ──
               │    │
               │    │
            ── +    +
                    │
                    │
        """
        if isinstance(generator, AlternatingSignMatrix):
            SVM = generator.to_six_vertex_model()
        elif isinstance(generator, (SquareIceModel.Element,
                                    SixVertexConfiguration)):
            SVM = generator
        else:  # Not ASM nor SVM
            try:
                SVM = AlternatingSignMatrix(generator).to_six_vertex_model()
            except (TypeError, ValueError):
                SVM = SixVertexModel(self._n, boundary_conditions='ice')(generator)
                SVM.to_alternating_sign_matrix()
        if len(SVM) != self._n:
            raise ValueError("invalid size")
        return self.element_class(self, SVM)

    Element = FullyPackedLoop

    def size(self):
        r"""
        Return the size of the matrices in ``self``.

        TESTS::

            sage: FPLs = FullyPackedLoops(4)
            sage: FPLs.size()
            4
        """
        return self._n

    def cardinality(self):
        r"""
        Return the cardinality of ``self``.

        The number of fully packed loops on  `n \times n` grid

        .. MATH::

            \prod_{k=0}^{n-1} \frac{(3k+1)!}{(n+k)!} = \frac{1! 4! 7! 10!
            \cdots (3n-2)!}{n! (n+1)! (n+2)! (n+3)! \cdots (2n-1)!}.

        EXAMPLES::

            sage: [AlternatingSignMatrices(n).cardinality() for n in range(10)]
            [1, 1, 2, 7, 42, 429, 7436, 218348, 10850216, 911835460]
        """
        return Integer(prod(factorial(3 * k + 1) / factorial(self._n + k)
                            for k in range(self._n)))

    def _an_element_(self):
        """
        Return an element of ``self``.

        EXAMPLES::

            sage: FPLs = FullyPackedLoops(3)
            sage: FPLs.an_element()
               │         │
               │         │
               +    + ── +
               │    │
               │    │
            ── +    +    + ──
                    │    │
                    │    │
               + ── +    +
               │         │
               │         │
        """
        # ASM = AlternatingSignMatrix(matrix.identity(self._n))
        # SVM = ASM.to_six_vertex_model()
        SVM = SixVertexModel(self._n, boundary_conditions='ice').an_element()
        return self.element_class(self, SVM)

    def _boundary(self, k):
        r"""
        Return the coordinates of the ``k``-th boundary.

        TESTS::

            sage: F = FullyPackedLoops(5)
            sage: [F._boundary(k) for k in range(10)] == F._boundaries()
            True
            sage: all(F._boundary_index(F._boundary(k)) == k for k in range(10))
            True

            sage: F = FullyPackedLoops(6)
            sage: [F._boundary(k) for k in range(12)] == F._boundaries()
            True
            sage: all(F._boundary_index(F._boundary(k)) == k for k in range(12))
            True
        """
        n = self._n
        n_LR = n//2 if n % 2 == 0 else (n+1) // 2
        n_TB = n//2 if n % 2 == 0 else (n-1) // 2
        if k < n_LR:
            return (-1, 2*k)
        k -= n_LR
        if k < n_TB:
            return (n % 2 + 2*k, n)
        k -= n_TB
        if k < n_LR:
            return (n, n - 1 - 2*k)
        k -= n_LR
        if k < n_TB:
            return (n - 1 - n % 2 - 2*k, -1)

    def _boundary_index(self, pos):
        r"""
        Return the index of the boundary at position ``pos``.

        TESTS::

            sage: F = FullyPackedLoops(5)
            sage: [F._boundary_index(b) for b in F._boundaries()]
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            sage: all(F._boundary(F._boundary_index(b)) == b for b in F._boundaries())
            True

            sage: F = FullyPackedLoops(6)
            sage: [F._boundary_index(b) for b in F._boundaries()]
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
            sage: all(F._boundary(F._boundary_index(b)) == b for b in F._boundaries())
            True
        """
        n = self._n
        i, j = pos
        if i == -1:
            return j//2
        if j == n:
            return (n + 1) // 2 + i // 2
        if i == n:
            return n + (n - j) // 2
        if j == -1:
            return 3 * n // 2 + (n - i) // 2

    def _boundaries(self):
        r"""
        Return the list of coordinates for the link in the boundaries.

        TESTS::

            sage: FullyPackedLoops(5)._boundaries()
            [(-1, 0), (-1, 2), (-1, 4), (1, 5), (3, 5),
             (5, 4),  (5, 2), (5, 0), (3, -1), (1, -1)]

            sage: FullyPackedLoops(6)._boundaries()
            [(-1, 0), (-1, 2), (-1, 4), (0, 6), (2, 6), (4, 6),
             (6, 5), (6, 3), (6, 1), (5, -1), (3, -1), (1, -1)]
        """
        n = self._n
        boundaries = []
        # left side: j = 0 mod 2
        boundaries.extend((-1, j) for j in range(0, n, 2))
        # top side: i = n mod 2
        boundaries.extend((i, n) for i in range(n % 2, n, 2))
        # right side: j = n+1 mod 2
        boundaries.extend((n, j) for j in range(n - 1, -1, -2))
        # bottom side: i = 1 mod 2
        boundaries.extend((i, -1) for i in range(n - 1 - n % 2, -1, -2))
        return boundaries



# distutils: libraries = flint
# distutils: depends = flint/mpf_vec.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    mpf * _mpf_vec_init(slong len, mp_limb_t prec) noexcept
    void _mpf_vec_clear(mpf * vec, slong len) noexcept
    void _mpf_vec_randtest(mpf * f, flint_rand_t state, slong len, flint_bitcnt_t bits) noexcept
    void _mpf_vec_zero(mpf * vec, slong len) noexcept
    void _mpf_vec_set(mpf * vec1, const mpf * vec2, slong len2) noexcept
    void _mpf_vec_set_fmpz_vec(mpf * appv, const fmpz * vec, slong len) noexcept
    bint _mpf_vec_equal(const mpf * vec1, const mpf * vec2, slong len) noexcept
    bint _mpf_vec_is_zero(const mpf * vec, slong len) noexcept
    bint _mpf_vec_approx_equal(const mpf * vec1, const mpf * vec2, slong len, flint_bitcnt_t bits) noexcept
    void _mpf_vec_add(mpf * res, const mpf * vec1, const mpf * vec2, slong len2) noexcept
    void _mpf_vec_sub(mpf * res, const mpf * vec1, const mpf * vec2, slong len2) noexcept
    void _mpf_vec_scalar_mul_mpf(mpf * res, const mpf * vec, slong len, mpf_t c) noexcept
    void _mpf_vec_scalar_mul_2exp(mpf * res, const mpf * vec, slong len, flint_bitcnt_t exp) noexcept
    void _mpf_vec_dot(mpf_t res, const mpf * vec1, const mpf * vec2, slong len2) noexcept
    void _mpf_vec_norm(mpf_t res, const mpf * vec, slong len) noexcept
    int _mpf_vec_dot2(mpf_t res, const mpf * vec1, const mpf * vec2, slong len2, flint_bitcnt_t prec) noexcept
    void _mpf_vec_norm2(mpf_t res, const mpf * vec, slong len, flint_bitcnt_t prec) noexcept



r"""
Homspaces between free modules

EXAMPLES:

We create `\mathrm{End}(\ZZ^2)` and compute a basis. ::

    sage: M = FreeModule(IntegerRing(), 2)
    sage: E = End(M)
    sage: B = E.basis()
    sage: len(B)
    4
    sage: B[0]
    Free module morphism defined by the matrix
    [1 0]
    [0 0]
    Domain:   Ambient free module of rank 2 over the principal ideal domain ...
    Codomain: Ambient free module of rank 2 over the principal ideal domain ...

We create `\mathrm{Hom}(\ZZ^3, \ZZ^2)` and compute a basis. ::

    sage: V3 = FreeModule(IntegerRing(), 3)
    sage: V2 = FreeModule(IntegerRing(), 2)
    sage: H = Hom(V3, V2)
    sage: H
    Set of Morphisms
     from Ambient free module of rank 3 over the principal ideal domain Integer Ring
       to Ambient free module of rank 2 over the principal ideal domain Integer Ring
       in Category of finite dimensional modules with basis over
          (Dedekind domains and euclidean domains
           and noetherian rings
           and infinite enumerated sets and metric spaces)
    sage: B = H.basis()
    sage: len(B)
    6
    sage: B[0]
    Free module morphism defined by the matrix
    [1 0]
    [0 0]
    [0 0]...

TESTS::

    sage: H = Hom(QQ^2, QQ^1)
    sage: loads(dumps(H)) == H
    True

See :issue:`5886`::

    sage: V = (ZZ^2).span_of_basis([[1,2],[3,4]])
    sage: V.hom([V.0, V.1])
    Free module morphism defined by the matrix
    [1 0]
    [0 1]...

See :issue:`13321`::

    sage: (GF(7)^2).hom([[20, 0], [0, 21]], ZZ^2)
    Traceback (most recent call last):
    ...
    TypeError: nontrivial morphisms require a coercion map from the base ring
    of the domain to the base ring of the codomain
"""
# ****************************************************************************
#  Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty
#    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License for more details; the full text
#  is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
import sage.categories.homset
from sage.structure.element import Matrix
from sage.matrix.constructor import matrix, identity_matrix
from sage.matrix.matrix_space import MatrixSpace
from sage.misc.cachefunc import cached_method


class FreeModuleHomspace(sage.categories.homset.HomsetWithBase):
    def __call__(self, A, **kwds):
        r"""
        INPUT:

        - ``A`` -- either a matrix or a list/tuple of images of generators,
          or a function returning elements of the codomain for elements of the domain
        - ``check`` -- boolean (default: ``True``)
        - the keyword ``side`` can be assigned the values ``'left'`` or
          ``'right'``. It corresponds to the side of vectors relative to the matrix.

        If A is a matrix, then it is the matrix of this linear
        transformation, with respect to the basis for the domain and
        codomain.  Thus the identity matrix always defines the
        identity morphism.

        EXAMPLES::

            sage: V = (ZZ^3).span_of_basis([[1,1,0],[1,0,2]])
            sage: H = V.Hom(V); H
            Set of Morphisms from ...
            sage: H([V.0,V.1])                    # indirect doctest
            Free module morphism defined by the matrix
            [1 0]
            [0 1]...
            sage: phi = H([V.1,V.0]); phi
            Free module morphism defined by the matrix
            [0 1]
            [1 0]...
            sage: phi(V.1) == V.0
            True
            sage: phi(V.0) == V.1
            True

        The following tests against a bug that was fixed in
        :issue:`9944`. The method ``zero()`` calls this hom space with
        a function, not with a matrix, and that case had previously
        not been taken care of::

            sage: V = span([[1/2,1,1],[3/2,2,1],[0,0,1]],ZZ)
            sage: V.Hom(V).zero()   # indirect doctest
            Free module morphism defined by the matrix
            [0 0 0]
            [0 0 0]
            [0 0 0]
            Domain:   Free module of degree 3 and rank 3 over Integer Ring
                      Echelon ...
            Codomain: Free module of degree 3 and rank 3 over Integer Ring
                      Echelon ...

        The following tests the bug fixed in :issue:`31818`. If there is no
        coercion between base rings, one can only define the zero morphism,
        as morphism of additive groups. Before one could for example use an
        integer matrix to define a morphism from the rational numbers to the
        integers. ::

            sage: V = QQ^2; W = ZZ^2; m = identity_matrix(2)
            sage: H = V.Hom(W); H(m)
            Traceback (most recent call last):
            ...
            TypeError: nontrivial morphisms require a coercion map from the base ring of the domain to the base ring of the codomain
            sage: n = zero_matrix(2);
            sage: h = H(n); h
            Free module morphism defined by the matrix
            [0 0]
            [0 0]
            Domain:   Vector space of dimension 2 over Rational Field
            Codomain: Ambient free module of rank 2 over the principal ideal domain Integer Ring
            sage: [h(v) for v in V.gens()]
            [(0, 0), (0, 0)]
        """
        from . import free_module_morphism
        side = kwds.get("side", "left")
        if not isinstance(A, Matrix):
            # Compute the matrix of the morphism that sends the
            # generators of the domain to the elements of A.
            C = self.codomain()
            try:
                if callable(A):
                    v = [C(A(g)) for g in self.domain().gens()]
                    A = matrix([C.coordinates(a) for a in v], ncols=C.rank())
                    if side == "right":
                        A = A.transpose()
                else:
                    v = [C(a) for a in A]
                    if side == "right":
                        A = matrix([C.coordinates(a) for a in v], ncols=C.rank()).transpose()
                    else:
                        A = matrix([C.coordinates(a) for a in v], ncols=C.rank())
            except TypeError:
                # Let us hope that FreeModuleMorphism knows to handle
                # that case
                pass
        if not self.codomain().base_ring().has_coerce_map_from(self.domain().base_ring()) and not A.is_zero():
            raise TypeError("nontrivial morphisms require a coercion map from the base ring of the domain to the base ring of the codomain")
        return free_module_morphism.FreeModuleMorphism(self, A, side)

    @cached_method
    def zero(self, side='left'):
        """
        INPUT:

        - ``side`` -- side of the vectors acted on by the matrix
          (default: ``'left'``)

        EXAMPLES::

            sage: E = ZZ^2
            sage: F = ZZ^3
            sage: H = Hom(E, F)
            sage: f = H.zero()
            sage: f
            Free module morphism defined by the matrix
            [0 0 0]
            [0 0 0]
            Domain:   Ambient free module of rank 2 over the principal ideal domain Integer Ring
            Codomain: Ambient free module of rank 3 over the principal ideal domain Integer Ring
            sage: f(E.an_element())
            (0, 0, 0)
            sage: f(E.an_element()) == F.zero()
            True
            sage: H.zero("right")
            Free module morphism defined as left-multiplication by the matrix
            [0 0]
            [0 0]
            [0 0]
            Domain:   Ambient free module of rank 2 over the principal ideal domain Integer Ring
            Codomain: Ambient free module of rank 3 over the principal ideal domain Integer Ring


        TESTS:

        We check that ``H.zero()`` is picklable::

            sage: loads(dumps(f.parent().zero()))
            Free module morphism defined by the matrix
            [0 0 0]
            [0 0 0]
            Domain:   Ambient free module of rank 2 over the principal ideal domain Integer Ring
            Codomain: Ambient free module of rank 3 over the principal ideal domain Integer Ring
        """
        return self(lambda x: self.codomain().zero(), side=side)

    @cached_method
    def _matrix_space(self, side='left'):
        """
        INPUT:

        - ``side`` -- side of the vectors acted on by the matrix
          (default: ``'left'``)

        Return underlying matrix space that contains the matrices that define
        the homomorphisms in this free module homspace.

        OUTPUT: matrix space

        EXAMPLES::

            sage: H = Hom(QQ^3, QQ^2)
            sage: H._matrix_space()
            Full MatrixSpace of 3 by 2 dense matrices over Rational Field
        """
        if side not in ["left", "right"]:
            raise ValueError("the side must be either 'left' or 'right'")
        R = self.codomain().base_ring()
        if side == "left":
            return MatrixSpace(R, self.domain().rank(), self.codomain().rank())
        if side == "right":
            return MatrixSpace(R, self.codomain().rank(), self.domain().rank())

    @cached_method
    def basis(self, side='left'):
        """
        Return a basis for this space of free module homomorphisms.

        INPUT:

        - ``side`` -- side of the vectors acted on by the matrix
          (default: ``'left'``)

        OUTPUT: tuple

        EXAMPLES::

            sage: H = Hom(ZZ^2, ZZ^1)
            sage: H.basis()
            (Free module morphism defined by the matrix
              [1]
              [0]
              Domain:   Ambient free module of rank 2 over the principal ideal domain ...
              Codomain: Ambient free module of rank 1 over the principal ideal domain ...,
             Free module morphism defined by the matrix
              [0]
              [1]
              Domain:   Ambient free module of rank 2 over the principal ideal domain ...
              Codomain: Ambient free module of rank 1 over the principal ideal domain ...)
            sage: H.basis("right")
            (Free module morphism defined as left-multiplication by the matrix
              [1 0]
              Domain:   Ambient free module of rank 2 over the principal ideal domain ...
              Codomain: Ambient free module of rank 1 over the principal ideal domain ...,
             Free module morphism defined as left-multiplication by the matrix
              [0 1]
              Domain:   Ambient free module of rank 2 over the principal ideal domain ...
              Codomain: Ambient free module of rank 1 over the principal ideal domain ...)
        """
        M = self._matrix_space(side)
        B = M.basis()
        return tuple([self(x, side=side) for x in B])

    def identity(self, side='left'):
        r"""
        Return identity morphism in an endomorphism ring.

        INPUT:

        - ``side`` -- side of the vectors acted on by the matrix
          (default: ``'left'``)

        EXAMPLES::

            sage: V = FreeModule(ZZ,5)
            sage: H = V.Hom(V)
            sage: H.identity()
            Free module morphism defined by the matrix
            [1 0 0 0 0]
            [0 1 0 0 0]
            [0 0 1 0 0]
            [0 0 0 1 0]
            [0 0 0 0 1]
            Domain:   Ambient free module of rank 5 over the principal ideal domain ...
            Codomain: Ambient free module of rank 5 over the principal ideal domain ...
        """
        if self.is_endomorphism_set():
            return self(identity_matrix(self.base_ring(), self.domain().rank()),
                        side=side)
        raise TypeError("Identity map only defined for endomorphisms. Try natural_map() instead.")



# sage.doctest: optional - numpy
"""
Hidden Markov Models -- Utility functions

AUTHOR:

- William Stein, 2010-03
"""

# ##########################################################################
#       Copyright (C) 2010 William Stein <wstein@gmail.com>
#  Distributed under the terms of the GNU General Public License (GPL) v2+.
#  The full text of the GPL is available at:
#                  https://www.gnu.org/licenses/
# ##########################################################################


from sage.structure.element import Matrix
from sage.misc.flatten import flatten


cdef class HMM_Util:
    """
    A class used in order to share cdef's methods between different files.
    """
    cpdef normalize_probability_TimeSeries(self, TimeSeries T, Py_ssize_t i, Py_ssize_t j):
        """
        This function is used internally by the Hidden Markov Models code.

        Replace entries of T[i:j] in place so that they are all
        nonnegative and sum to 1.  Negative entries are replaced by 0 and
        T[i:j] is then rescaled to ensure that the sum of the entries in
        each row is equal to 1.  If all entries are 0, replace them
        by 1/(j-i).

        INPUT:

        - ``T`` -- a :class:`TimeSeries`
        - ``i`` -- nonnegative integer
        - ``j`` -- nonnegative integer

        OUTPUT: ``T`` is modified

        EXAMPLES::

            sage: import sage.stats.hmm.util
            sage: T = stats.TimeSeries([.1, .3, .7, .5])
            sage: u = sage.stats.hmm.util.HMM_Util()
            sage: u.normalize_probability_TimeSeries(T,0,3)
            sage: T
            [0.0909, 0.2727, 0.6364, 0.5000]
            sage: u.normalize_probability_TimeSeries(T,0,4)
            sage: T
            [0.0606, 0.1818, 0.4242, 0.3333]
            sage: abs(T.sum()-1) < 1e-8    # might not exactly equal 1 due to rounding
            True
        """
        # One single bounds check only
        if i < 0 or j < 0 or i > T._length or j > T._length:
            raise IndexError

        if j-i <= 0:
            # Nothing to do
            return

        cdef Py_ssize_t k

        # Replace negative entries by 0, summing entries
        cdef double s = 0, t
        for k in range(i,j):
            if T._values[k] < 0:
                T._values[k] = 0
            else:
                s += T._values[k]

        if s == 0:
            # If sum is 0, make all entries 1/(j-i).
            t = 1.0/(j-i)
            for k in range(i,j):
                T._values[k] = t
        else:
            # Normalise so sum is 1.
            for k in range(i, j):
                T._values[k] /= s

    cpdef TimeSeries initial_probs_to_TimeSeries(self, pi, bint normalize):
        """
        This function is used internally by the __init__ methods of
        various Hidden Markov Models.

        INPUT:

        - ``pi`` -- vector, list, or :class:`TimeSeries`
        - ``normalize`` -- if ``True``, replace negative entries by 0 and
          rescale to ensure that the sum of the entries in each row is
          equal to 1.  If the sum of the entries in a row is 0, replace them
          all by `1/N`.

        OUTPUT: a :class:`TimeSeries` of length `N`

        EXAMPLES::

            sage: import sage.stats.hmm.util
            sage: u = sage.stats.hmm.util.HMM_Util()
            sage: u.initial_probs_to_TimeSeries([0.1,0.2,0.9], True)
            [0.0833, 0.1667, 0.7500]
            sage: u.initial_probs_to_TimeSeries([0.1,0.2,0.9], False)
            [0.1000, 0.2000, 0.9000]
        """
        cdef TimeSeries T
        if isinstance(pi, TimeSeries):
            T = pi
        else:
            if not isinstance(pi, list):
                pi = list(pi)
            T = TimeSeries(pi)
        if normalize:
            # Now normalize
            self.normalize_probability_TimeSeries(T, 0, T._length)
        return T

    cpdef TimeSeries state_matrix_to_TimeSeries(self, A, int N, bint normalize):
        """
        This function is used internally by the ``__init__`` methods of
        Hidden Markov Models to make a transition matrix from ``A``.

        INPUT:

        - ``A`` -- matrix, list, list of lists, or :class:`TimeSeries`
        - ``N`` -- number of states
        - ``normalize`` -- if ``True``, replace negative entries by 0 and
          rescale to ensure that the sum of the entries in each row is
          equal to 1.  If the sum of the entries in a row is 0, replace them
          all by `1/N`.

        OUTPUT: a :class:`TimeSeries`

        EXAMPLES::

            sage: import sage.stats.hmm.util
            sage: u = sage.stats.hmm.util.HMM_Util()
            sage: u.state_matrix_to_TimeSeries([[.1,.7],[3/7,4/7]], 2, True)
            [0.1250, 0.8750, 0.4286, 0.5714]
            sage: u.state_matrix_to_TimeSeries([[.1,.7],[3/7,4/7]], 2, False)
            [0.1000, 0.7000, 0.4286, 0.5714]
        """
        cdef TimeSeries T
        if isinstance(A, TimeSeries):
            T = A
        elif isinstance(A, Matrix):
            T = TimeSeries(A.list())
        elif isinstance(A, list):
            T = TimeSeries(flatten(A))
        else:
            T = TimeSeries(A)
        cdef Py_ssize_t i
        if normalize:
            # Set to 0 negative rows and make sure sum of entries in each
            # row is 1.
            if len(T) != N*N:
                raise ValueError("number of entries of transition matrix A must be the square of the number of entries of pi")
            for i in range(N):
                self.normalize_probability_TimeSeries(T, i*N, (i+1)*N)
        return T
